{}
{                                                       }
{      Virtual Pascal Examples  Version 2.1             }
{      TOUCH command line utility.                      }
{      }
{      Copyright (C) 2000 vpascal.com                   }
{                                                       }
{}
{$I-,D+,Delphi-,Use32+}
{$M 500000}

{ Changes the date and time of selected files to the date }
{ and time at which the program is run.                   }

program Touch;

{&PMTYPE VIO}

uses Dos;

var
  Time: Longint;
  I: Integer;
  ParamStrI: String;

Const
  RecurseSubdirs : Boolean = False;
  NumberOfProcessedFiles : Word = 0;
  FoundProblems : Boolean = False;
  IgnoreLocked: Boolean = False;

{ Reports an error }

procedure Error(const ErrStr: String);
begin
  WriteLn('**Error**  ', ErrStr);
  FoundProblems := True;
end;

{ Returns current date & time in the packed file format }

function GetDateTime: Longint;
var
  DT: DateTime;
  DayOfWeek,Sec100: Word;
  Time: Longint;
begin
  GetDate(DT.Year, DT.Month, DT.Day, DayOfWeek);
  GetTime(DT.Hour, DT.Min  , DT.Sec, Sec100);
  PackTime(DT, Time);
  GetDateTime := Time;
end;


procedure SetTimeValue(var w: Word;var s: String;const MinVal, MaxVal: Word);
var
  temp: Word;
begin
  if s='' then Exit;
  temp:=0;
  while (s<>'') and (not (s[1] in [':','/'])) do
    begin
      if s[1] in ['0'..'9'] then
        temp:=10*temp+Ord(s[1])-Ord('0')
      else
        Exit;
      Delete(s,1,1);
    end;
  Delete(s,1,1);

  if MinVal=1980 then
    begin
      Inc(temp,1900);
      if temp < 1980 then
        Inc(temp,100); // 19xx/20xx
    end;
  if (temp>=MinVal) and (temp<=MaxVal) then
    w := temp;
end;

function SetupDate(s:string):boolean;
var
  DT: DateTime;
begin
  if  (Length(s) >= Length('-dm/d/y'))
    and (Length(s) <= Length('-dmm/dd/yyyy'))
    and (s[1]='-') and (s[2]='d') and (Pos('/',s) <> 0) then
    with DT do
      begin
        Delete(s,1,Length('-d'));
        UnpackTime(Time, DT);
        SetTimeValue(Month,s,   0,  15);
        SetTimeValue(Day  ,s,   0,  31);
        SetTimeValue(Year ,s,1980,2079);
        PackTime(DT, Time);
        SetupDate := True;
      end
  else
    SetupDate := False;
end;

function SetupTime(s:string):boolean;
var
  DT: DateTime;
begin
  if  (Length(s) >= Length('-th:m'))
    and (Length(s) <= Length('-thh:mm:ss'))
    and (s[1]='-') and (s[2]='t') and (Pos(':',s)<>0) then
    with DT do
      begin
        Delete(s,1,Length('-t'));
        UnpackTime(Time, DT);
        SetTimeValue(Hour ,s,   0,  31);
        SetTimeValue(Min  ,s,   0,  63);
        SetTimeValue(Sec  ,s,   0,  62);
        PackTime(DT, Time);
        SetupTime:=true;
      end
  else
    SetupTime:=false;
end;

{ Touches file(s) specified by a given file name (wildcard) }

procedure DoTouch(const Wild: String);
var
  SR: SearchRec;
  F: File;
  FName: PathStr;
  Dir: DirStr;
  Name: NameStr;
  Ext: ExtStr;
begin
  FSplit(Wild, Dir, Name, Ext);
  FindFirst(Wild, AnyFile, SR);
  if DosError <> 0 then
    Error('Cannot locate file ' + Wild)
  else
    while DosError = 0 do
      begin
        if (SR.Attr and Directory)=Directory then
          begin // Directory ..
            if RecurseSubdirs and (SR.Name <> '.') and (SR.Name <> '..') then
              DoTouch(Dir+SR.Name+'\*');
          end
        else
          begin // File ..
            FName := Dir + SR.Name;
            Assign(F, FName);
            { OS/2: Write only, deny read-write, fail on error }
            { DOS: Write only }
            FileMode := {$IFDEF OS2} $2011 {$ELSE} $21 {$ENDIF};
            Reset(F);
            if (IOResult <> 0) and not IgnoreLocked then
              Error('Unable to open file ' + FName);
            SetFTime(F, Time);
            if (DosError <> 0) and not IgnoreLocked then
              Error('Unable to change file date and time: ' + FName);
            Close(F); InOutRes := 0;
            Inc(NumberOfProcessedFiles);
          end;
    FindNext(SR);
  end;
{$IFDEF VIRTUALPASCAL}
  FindClose(SR);
{$ENDIF}
end;

begin
  {$IFDEF LINUX}
  FileSystem := fsDos;
  {$ENDIF}

  WriteLn('Virtual Pascal Touch  Version 2.1 Copyright (C) 2000 vpascal.com');
  if ParamCount = 0 then
  begin
    WriteLn('Syntax: TOUCH [-thh:mm:ss] [-dmm/dd/yy] [-s] [-i] Arg1 [Arg2..Argn]');
    WriteLn('where ArgX are file names or wildcards.');
    Writeln(' -s: Touch subdirectories');
    Writeln(' -i: Ignore locked files');
    Halt(1);
  end;
  Time := GetDateTime;
  for I := 1 to ParamCount do
    begin
      ParamStrI:=ParamStr(I);
      if ParamStrI = '-s' then
        RecurseSubdirs := Not RecurseSubdirs
      else if ParamStrI = '-i' then
        IgnoreLocked := True
      else if SetupDate(ParamStrI) then
      else if SetupTime(ParamStrI) then
      else
        DoTouch(ParamStrI);
    end;

  if NumberOfProcessedFiles = 0 then
    FoundProblems := True;

  if FoundProblems then
    Halt(1)
end.

