// |---------------------------------------------------------|
// |                                                         |
// |      Virtual Pascal Runtime Library.  Version 2.1.      |
// |      Unit Test for Objects                              |
// |      ---------------------------------------------------|
// |      Copyright (C) 2003 vpascal.com                     |
// |---------------------------------------------------------|

unit ObjTest;
{&Use32+}

interface

uses
  UnitTest;

type
  tTestObjectConstruction = class(tTest)
    function Run: Boolean; override;
  end;

implementation

type
  PObject1 = ^TObject1;
  TObject1 = object
  protected
    function TestFunc(const S: string): String; virtual;
  public
    constructor Init1;
    constructor Init2;
    destructor Done; virtual;
  end;

  PObject2 = ^TObject2;
  TObject2 = object(TObject1)
  protected
    function TestFunc(const S: string): String; virtual;
  end;

  PObject3 = ^TObject3;
  TObject3 = object(TObject2)
  protected
    function TestFunc(const S: string): String; virtual;
  public
    constructor Init1;
    constructor Init2;
    constructor Init3;
  end;

constructor TObject1.Init1;
begin
  Init2;
end;

constructor TObject1.Init2;
begin
  // Nothing
end;

destructor TObject1.Done;
begin
end;

function  TObject1.TestFunc(const S: string): String;
begin
  Result := 'Object1 '+S;
end;

function  TObject2.TestFunc(const S: string): String;
begin
  Result := 'Object2 '+S;
end;

constructor TObject3.Init1;
begin
  Init2;
end;

constructor TObject3.Init2;
begin
  inherited Init1;
end;

constructor TObject3.Init3;
begin
  Init1;
end;

function  TObject3.TestFunc(const S: string): String;
begin
  Result := 'Object3 '+S;
end;

// tTestObjectConstruction

function tTestObjectConstruction.Run: Boolean;
var
  P1: PObject1;
  P2: PObject2;
  P3: PObject3;

  T1: TObject1;
  T2: TObject2;
  T3: TObject3;

begin
  // Object1 instances
  P1 := New(PObject1, Init1);
  Result := Test( P1^.TestFunc('Init1') = 'Object1 Init1', 'Test 1');
  Dispose(P1, Done);

  P1 := New(PObject1, Init2);
  Result := Result and Test( P1^.TestFunc('Init2') = 'Object1 Init2', 'Test 2');
  Dispose(P1, Done);

  // Object2 instances
{ Bug #81:
  P1 := New(PObject2, Init1);
  Result := Result and Test( P1^.TestFunc('Init1') = 'Object2 Init1', 'Test 3');
  Dispose(P1, Done);

  P2 := New(PObject2, Init1);
  Result := Result and Test( P1^.TestFunc('Init1') = 'Object2 Init1', 'Test 4' );
  Dispose(P2, Done);
}
  P1 := New(PObject2, Init2);
  Result := Result and Test( P1^.TestFunc('Init2') = 'Object2 Init2', 'Test 5' );
  Dispose(P1, Done);

  P2 := New(PObject2, Init2);
  Result := Result and Test( P1^.TestFunc('Init2') = 'Object2 Init2', 'Test 6' );
  Dispose(P2, Done);

  // Object3 instances
{ Bug #81:
  P1 := New(PObject3, Init1);
  Result := Result and Test( P1^.TestFunc('Init1') = 'Object3 Init1', 'Test 7');
  Dispose(P1, Done);

  P2 := New(PObject3, Init1);
  Result := Result and Test( P1^.TestFunc('Init1') = 'Object3 Init1', 'Test 8' );
  Dispose(P2, Done);

  P3 := New(PObject3, Init1);
  Result := Result and Test( P1^.TestFunc('Init1') = 'Object3 Init1', 'Test 9' );
  Dispose(P2, Done);

  P1 := New(PObject3, Init2);
  Result := Result and Test( P1^.TestFunc('Init2') = 'Object3 Init2', 'Test 10' );
  Dispose(P1, Done);

  P2 := New(PObject3, Init2);
  Result := Result and Test( P1^.TestFunc('Init2') = 'Object3 Init2', 'Test 11' );
  Dispose(P2, Done);

  P3 := New(PObject3, Init2);
  Result := Result and Test( P1^.TestFunc('Init2') = 'Object3 Init2', 'Test 12' );
  Dispose(P3, Done);

  P1 := New(PObject3, Init3);
  Result := Result and Test( P1^.TestFunc('Init3') = 'Object3 Init3', 'Test 13' );
  Dispose(P1, Done);
}  
  T1.Init1;
  Result := Result and Test( T1.TestFunc('Init1') = 'Object1 Init1', 'Test 14' );
  T1.Done;

  T2.Init2;
  Result := Result and Test( T2.TestFunc('Init2') = 'Object2 Init2', 'Test 15' );
  T1.Done;
{ Bug #81:
  T3.Init3;
  Result := Result and Test( T3.TestFunc('Init3') = 'Object3 Init3', 'Test 16' );
  T1.Done;
}
end;

initialization
  RegisterTests( [tTestObjectConstruction] );
end.

