// |---------------------------------------------------------|
// |                                                         |
// |      Virtual Pascal Runtime Library.  Version 2.1.      |
// |      Unit Test for String Handling Routines             |
// |      ---------------------------------------------------|
// |      Copyright (C) 2003 vpascal.com                     |
// |---------------------------------------------------------|

unit StrTest;
{&Use32+}

interface

uses
  Classes, SysUtils, UnitTest, Math;

type
  // Test cases adopted from Free Pascal
  tTestStringsFpkShort = class(tTest)
    function Run: Boolean; override;
    function Name: String; override;
  end;

  tTestStringsFpkAnsi = class(tTest)
    function Run: Boolean; override;
    function Name: String; override;
  end;

  tTestCharArrays = class(tTest)
    function Run: Boolean; override;
  end;

  tTestString4Fpk = class(tTest)
  private
    function TestInitFinal: Boolean;
  public
    function Run: Boolean; override;
  end;

  tTestStringParams = class(tTest)
  private
    function TestVarParam (Var Sv : AnsiString): Boolean;
    function TestValParam (S : AnsiString): Boolean;
    function TestConstParam (Const Sc : AnsiString): Boolean;
  public
    function Run: Boolean; override;
  end;

  tTestStringComparisons = class(tTest)
    function Run: Boolean; override;
  end;

  // Systematic test of Val and Str
  tTestStringsValStr = class(tTest)
  private
    function TestStr(const _S: String): Boolean;
    function TestHexStr(const _S: String): Boolean;
  public
    function Run: Boolean; override;
  end;

  tTestPChars = class(tTest)
  public
    function Run: Boolean; override;
  end;

  tTestAnsiStrings = class(tTest)
  public
    function Run: Boolean; override;
  end;

implementation

uses
  Strings;

function tTestStringsFpkShort.Name: String;
begin
  Result := 'String test 1 from FreePascal test suite (Short)';
end;

function tTestStringsFpkShort.Run: Boolean;
Const
  TestStr: string[22]='HELLO, THIS IS A TEST ';
var
  I : Integer;
  U : String[1];
  Q : String[100];
  S : String[55];
  T : String[60];
  V : ShortString;
begin
  Result := False;
  T := 'THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG 1234567890';
  Insert(T, T, 1);
  Delete(T, 55, 54);
  S := Copy(T, -5, 2);
  U := Copy(T, 7, 4);
  S := S + U;
  if not Test((s = 'THI') and (u = 'I'), 'Stage 1') then
    exit;

  Q := Copy(T, 32, 70);
  Delete(Q, 2, 1);
  Delete(Q, 100, 2);
  Delete(Q, 3, -4);
  Delete(Q, 3, 10);
  if not Test(q = 'TE1234567890', 'Stage 2: Delete') then
    exit;

  I := Pos('S', T);
  Insert(Copy(T,I,200),Q,3);
  Delete(Q, 4, 6);
  S := S + T [25];
  if not Test(s = 'THIS', 'Stage 3') then
    exit;

  S := S + Copy(S, 3, -5) + Copy(S, 3, 2);  {'THISIS'}
  V := T;
  // Test NOP:
  Delete(V, -10, 47);
  Delete(V, 1, -10);
  if not Test(v = t, 'Delete w/invalid parms') then
    exit;

  Delete(V, 1, 36);
  if not Test(v = 'AZY DOG 1234567890', 'Stage 4') then
    exit;

  if (Copy(V, -7, -1)='') and(Pos('DOG', V)=5) then {TRUE}
    Insert(V, S, 200);
  U := Copy(T, 44, 40);
  Insert(U, S, 5);
  I := Pos('ZY', S);
  Delete(S, I, -5);        {'THIS ISAZY DOG 1234567890'}
  if not Test(s = 'THIS ISAZY DOG 1234567890', 'Stage 5') then
    exit;

  Insert(Copy(S,5,1),S,8);
  Delete(S, 10, 16);
  if S[Length(S)]='0' then
    S := S + Q;
  V := Copy(S, Length(S) - 19, 10);
  if not Test(v = '1234567890', 'Stage 6') then
    exit;

  if V=Copy(S, Length(S) - 9, 10) then
    Delete(S, 15, 3 * Length(V)+2);
  if not Test(s = 'THIS IS A0TEST', 'Stage 7') then
    exit;

  Insert('', S, 0);
  Insert(Copy(S,5,1),S,11);
  Insert('HELLO', S, -4);
  Insert(',', S, 6);
  Delete(S, Pos('TEST', S) - 2, 1);
  if not Test(s = 'HELLO,THIS IS A TEST', 'Stage 8') then
    exit;

  Delete(Q, 1, 32767);
  Q := Q + ' ';
  Insert(Q, S, 7);
  Insert(Q, S, 255);
  Result := Test(
    (s = TestStr) and
    (Q = ' ') and
    (V = '1234567890') and
    (T = 'THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG 1234567890'),
    'Final check');
end;

function tTestStringsFpkAnsi.Name: String;
begin
  Result := 'String test 1 from FreePascal test suite (Ansi)';
end;

function tTestStringsFpkAnsi.Run: Boolean;
Const
  TestStr: String[22]='HELLO, THIS IS A TEST ';
var
  I : Integer;
  U : String[1];
  Q : AnsiString;
  S : AnsiString;
  T : AnsiString;
  V : AnsiString;
begin
  Result := False;
  T := 'THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG 1234567890';
  Insert(T, T, 1);
  Delete(T, 55, 54);
  S := Copy(T, -5, 2);
  U := Copy(T, 7, 4);
  S := S + U;
  if not Test((s = 'THI') and (u = 'I'), 'Stage 1') then
    exit;

  Q := Copy(T, 32, 70);
  Delete(Q, 2, 1);
  Delete(Q, 100, 2);
  Delete(Q, 3, -4);
  Delete(Q, 3, 10);
  if not Test(q = 'TE1234567890', 'Stage 2: Delete') then
    exit;

  I := Pos('S', T);
  Insert(Copy(T,I,200),Q,3);
  Delete(Q, 4, 6);
  S := S + T [25];
  if not Test(s = 'THIS', 'Stage 3') then
    exit;

  S := S + Copy(S, 3, -5) + Copy(S, 3, 2);  {'THISIS'}
  V := T;
  // Test NOP:
  Delete(V, -10, 47);
  Delete(V, 1, -10);
  if not Test(v = t, 'Delete w/invalid parms') then
    exit;

  Delete(V, 1, 36);
  if not Test(v = 'AZY DOG 1234567890', 'Stage 4') then
    exit;

  if (Copy(V, -7, -1)='') and(Pos('DOG', V)=5) then {TRUE}
    Insert(V, S, 200);
  U := Copy(T, 44, 40);
  Insert(U, S, 5);
  I := Pos('ZY', S);
  Delete(S, I, -5);        {'THIS ISAZY DOG 1234567890'}
  if not Test(s = 'THIS ISAZY DOG 1234567890', 'Stage 5') then
    exit;

  Insert(Copy(S,5,1),S,8);
  Delete(S, 10, 16);
  if S[Length(S)]='0' then
    S := S + Q;
  V := Copy(S, Length(S) - 19, 10);
  if not Test(v = '1234567890', 'Stage 6') then
    exit;

  if V=Copy(S, Length(S) - 9, 10) then
    Delete(S, 15, 3 * Length(V)+2);
  if not Test(s = 'THIS IS A0TEST', 'Stage 7') then
    exit;

  Insert('', S, 0);
  Insert(Copy(S,5,1),S,11);
  Insert('HELLO', S, -4);
  Insert(',', S, 6);
  Delete(S, Pos('TEST', S) - 2, 1);
  if not Test(s = 'HELLO,THIS IS A TEST', 'Stage 8') then
    exit;

  Delete(Q, 1, 32767);
  Q := Q + ' ';
  Insert(Q, S, 7);
  Insert(Q, S, 255);
  Result := Test(
    (s = TestStr) and
    (Q = ' ') and
    (V = '1234567890') and
    (T = 'THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG 1234567890'),
    'Final check');
end;

function tTestCharArrays.Run: Boolean;
var
  a : array[1..10,1..10,1..5] of char;
  i,j,k,l : integer;

begin
  for i := 1 to 10 do
    a[i,i] := 'Hello';
  i := 1;
  j := 2;
  k := 3;
  l := 4;
  { test register allocation }
  Result := Test( (a[i,i] = 'Hello') and (a[j,j] = 'Hello') and
    (a[k,k] = 'Hello') and (a[l,l] = 'Hello'), 'Char array' );
  for i := 1 to 10 do
    Result := Result and Test(a[i,i] = 'Hello', 'Array index');
end;

// tTestString4Fpk - from FPK's tstring4.pp

function tTestString4Fpk.TestInitFinal: Boolean;
type
  ARec = record
    FirstName, LastName : AnsiString;
  end;
  AnArray = Array [1..10] of AnsiString;

var
  S   : AnsiString;
  AR  : Arec;
  AAR : AnArray;
  I   : longint;

begin
  S := 'This is an ansistring!';
  Result := Test(Pointer(AR.FirstNAme) = nil, 'AR.FirstName not OK');
  for I := 1 to 10 do
    Result := Result and Test(Pointer(AAR[I])=nil, Format('Array (%d) NOT ok', [i]) );
  AR.FirstName := 'Napoleon';
  AR.LastName  := 'Bonaparte';
  for I := 1 to 10 do
    AAR[I]:='Yet another AnsiString';
end;

function tTestString4Fpk.Run: Boolean;
begin
  Result := TestInitFinal;
end;

// tTestStringParams

function tTestStringParams.TestVarParam(Var Sv : AnsiString): Boolean;
var
  LS : AnsiString;

begin
  Result := Test(Sv = 'Test String', 'Var parameter = '+Sv);
  Sv := 'Changed Var Parameter';
  Ls := Sv;
  Result := Result and Test(Sv = Ls, 'Assigning LS to SV');
end;

function tTestStringParams.TestValParam(S : AnsiString): Boolean;
var
  LS : AnsiString;

begin
  Result := Test(S = 'Test String', 'Value parameter = '+s);
  S := 'Changed Value Parameter';
  Result := Result and Test(S = 'Changed Value Parameter', 'Changed value parameter = '+s);
  LS := S;
  Result := Result and Test(LS = S, 'Assigning LS to S');
end;

function tTestStringParams.TestConstParam (Const Sc : AnsiString): Boolean;
var
  LS : AnsiString;

begin
  Result := Test(Sc = 'Test String', 'Const parameter = '+sc);
  Ls := Sc;
  Result := Result and Test(Ls = Sc, 'Const parameter assignment');
end;

function tTestStringParams.Run: Boolean;
var
  S : AnsiString;
begin
  S := 'Test String';
  Result := TestValParam(s);
  Result := Result and TestConstParam(s);
  Result := Result and TestVarParam(s);
  Result := Result and Test(s = 'Changed Var Parameter', 'Var param return value');
end;

// tTestStringComparisons

function tTestStringComparisons.Run: Boolean;
const
  S1 : AnsiString = 'Teststring 1';
  S2 : AnsiString = 'Teststring 1';
  S3 : AnsiString = 'Teststring 2';
  S4 : AnsiString = '';
  PC : Pchar = 'Teststring 1';

var
  S,T : AnsiString;
  ss : Shortstring;

begin
  Result := Test(S1 = S2, 'S1 <> S2');
  Result := Result and Test(S4 = '', 'S4 is not empty');
  Result := Result and not Test(S4 = 'Non-empty', 'S4 is not-empty');
  Result := Result and Test(S3 = 'Teststring 2', 'S3 <> S2');
  Result := Result and Test(S2 <> S3, 'S2 = S3');
  Result := Result and Test(S3 > S2, 'S3 is not > S2');
  Result := Result and Test(S1 < S3, 'S1 is not < S3');
  S := S2;
  T := S;
  Result := Result and Test(S = T, 'S <> T');
  SS := 'Teststring 1';
  Result := Result and Test(S = SS, 'S <> SS');
  Result := Result and Test(S = PC, 'S <> PC');
end;

// tTestStringsValStr

function tTestStringsValStr.TestStr(const _S: String): Boolean;
var
  Value: Longint;
  Err: Longint;
  S: String;
begin
  Result := False;
  Val(_s, Value, Err);
  if Err <> 0 then
    begin
      FailReason := Format('Error in Val(%s): %d', [_S, Err]);
      exit;
    end;
  Str(Value, S);
  if S <> _S then
    begin
      FailReason := Format('Error in Str(%d) = %s', [Value, S]);
      exit;
    end;
  Result := True;
end;

function tTestStringsValStr.TestHexStr(const _S: String): Boolean;
var
  Value: Longint;
  Err: Longint;
  S: String;
begin
  Result := False;
  Val(_s, Value, Err);
  if Err <> 0 then
    begin
      FailReason := Format('Error in Val(%s): Error at pos. %d', [_S, Err]);
      exit;
    end;
  if (_S[1] <> '-') and (Format('$%x', [Value]) <> _S) then
    begin
      FailReason := Format('Hex Val error: %s <> $%x', [_S, Value]);
      exit;
    end;
  Str(Value, S);
  Result := TestStr(S);
end;

function tTestStringsValStr.Run: Boolean;
begin
  Result := TestStr( '0' );
  Result := Result and TestStr( '-100' );
  Result := Result and TestStr( '100' );
  Result := Result and TestStr( '2147483647' );
  Result := Result and TestStr( '-2147483648' );
  Result := Result and TestHexStr( '$0' );
  Result := Result and TestHexStr( '$80000000' );
  Result := Result and TestHexStr( '$7FFFFFFF' );
  Result := Result and TestHexStr( '-$80000000' );
  Result := Result and TestHexStr( '-$7FFFFFFF' );
  Result := Result and TestHexStr( '$FFFFFFFF' );
  Result := Result and TestHexStr( '-$FFFFFFFF' );
  // Also test cases that *should* fail
  Result := Result and not TestHexStr( '$100000000' );
  Result := Result and not TestHexStr( '-$100000000' );
end;

// tTestPChars
function tTestPChars.Run: Boolean;
const
  VeryLong: PChar =
    '1234567890123456789012345678901234567890123456789012345678901234567890' +
    '1234567890123456789012345678901234567890123456789012345678901234567890' +
    '1234567890123456789012345678901234567890123456789012345678901234567890' +
    '1234567890123456789012345678901234567890123456789012345678901234567890' +
    '1234567890123456789012345678901234567890123456789012345678901234567890';
var
  s: AnsiString;
  ss: ShortString;
  i, t, l: Integer;
  p, q: PChar;

begin
  // Test StrLen
  Result := Test( StrLen('123') = 3, 'StrLen 3' );
  Result := Result and Test( StrLen('1234') = 4, 'StrLen 4' );
  s := '';
  for i := 1 to 100 do
    begin
      s := s + 'x';
      Result := Result and Test( StrLen(PChar(s)) = Length(s), 'Len' );
    end;

  // Test StrCopy/StrEnd
  GetMem(p, 100);
  StrCopy( p, 'Virtual Pascal' );
  q := StrEnd( p );
  StrCopy( q, ' v2.1' );
  s := p;
  Result := Result and Test( s = 'Virtual Pascal v2.1', 'StrCopy/StrEnd' );

  // Test StrScan
  q := StrScan(p, '2'); s := q;
  Result := Result and Test( s = '2.1', 'StrScan 2' );
  q := StrScan(p, 'a'); s := q;
  Result := Result and Test( s = 'al Pascal v2.1', 'StrScan a' );
  q := StrScan(p, '8');
  Result := Result and Test( q = nil, 'StrScan 8' );

  // Test StrRScan
  q := StrRScan(p, '2');
  s := q;
  Result := Result and Test( s = '2.1', 'StrRScan 2' );
  q := StrRScan(p, 'a');
  s := q;
  Result := Result and Test( s = 'al v2.1', 'StrRScan a' );
  q := StrRScan(p, '8');
  Result := Result and Test( q = nil, 'StrRScan 8' );

  // Test StrPas
  ss := StrPas(p);
  Result := Result and Test( ss = 'Virtual Pascal v2.1', 'StrPas' );
  ss := StrPas( VeryLong );
  Result := Result and Test( Length(ss) = 255, 'StrPas Long' );

  FreeMem( p );
end;

// Ansi string tests

// >> Currently only tests simple assignment and SetLength

function tTestAnsiStrings.Run: Boolean;
var
  S: AnsiString;
begin
  S := 'Hello';
  Result := Test( S = 'Hello', 'Assignment' );
  SetLength(S, 2);
  Result := Result and Test( S = 'He', 'SetLength(2)' );
  SetLength(S, -7);
  Result := Result and Test( S = '', 'SetLength(-7)' );
end;

initialization
  RegisterTests( [tTestStringsFpkShort, tTestStringsFpkAnsi, tTestString4Fpk,
    tTestStringComparisons] );
  RegisterTests( [tTestStringParams, tTestCharArrays, tTestStringsValStr,
    tTestPChars, tTestAnsiStrings] );
end.

