{************************************************}
{                                                }
{   Chess - Shared DLL Example                   }
{   CHESS.DLL Task managment.                    }
{   Copyright (c) 1992 by Borland International  }
{                                                }
{************************************************}

{ This unit allocates and manages a separate stack that is used to
  preform the game searches.   It is *very* important that, when running
  under Windows, Windows calls are made while in the child task.  Many
  Windows calls use the stack to determine the task handle and the stack
  must be the DLL's client's stack or Windows will become confused. }

unit TaskMgr;

{$R-,Q-,G+,S-,W-}

interface

uses Use32;

type
  { Stores the state of the stack   }
  TTaskInfo = record
    SPtr, SBtm, Size: Word;
    SavedEbx,SavedEsi,SavedEdi: Longint;
  end;

  TSpawnProc = procedure;


{ Allocate the stack used by the game task }
procedure AllocateTask(Size: Word);

{ Dispose of the stack allocated by AllocateTask }
procedure DisposeTask;

{ Spawn a mini process which will call the given function upon start-up. }
procedure Spawn(Proc: TSpawnProc);

{ Suspend the current task waiting for a message from another task.  The
  return value of this function is the Msg paramter of the Message call
  perfomed by the other task. Each task will alternate (coopertivly
  multitask) calling Message. }
function Message(Msg: Word): Word;

implementation

{$IFDEF WINDOWS}
uses GameRec, OMemory;
{$ELSE}
uses GameRec, Memory;
{$ENDIF}

procedure AllocateTask(Size: Word);
var
  P : Pointer;
begin
  P := MemAlloc(Size);
  if P = nil then Exit;
  with CC do
  begin
    GameStack.SBtm := Longint(P);
    GameStack.Size := Size;
    GameStack.SPtr := GameStack.SBtm + Size - 4;
    AppStack.SBtm  := SPtr and $FFFF0000;
  end;
end;


procedure DisposeTask;
begin
  with CC.GameStack do
    if SBtm <> 0 then FreeMem(Pointer(SBtm));
  FillChar(CC.GameStack, 0, SizeOf(CC.GameStack));
end;

 { TerminateTask kills the game stack and RETurns to the ret addr  }
 { already on App stack, presumeably, the last caller of Resume.   }
 { The Game stack only shares a common starting point with the App }
 { stack - after that, it runs to completion without trying to     }
 { splice its execution back into the App stack.                   }

procedure TerminateTask(Dummy: Word); far; assembler;
asm
        { Switch to the App's stack }
        mov     esp,CC.AppStack.SPtr
        mov     ebp,esp
end;

function Message(Msg: Word): Word; assembler;
asm
        mov     eax,Msg
        mov     edx,esp
        and     edx,0FFFF0000h
        cmp     edx,CC.AppStack.SBtm
        jne     @@1

        { Switch to game stack }
        mov     CC.AppStack.SPtr,esp
        mov     CC.AppStack.SavedEbx,ebx
        mov     CC.AppStack.SavedEsi,esi
        mov     CC.AppStack.SavedEdi,edi

        mov     esp,CC.GameStack.SPtr   { Now on the new stack }
        mov     ebx,CC.GameStack.SavedEbx
        mov     esi,CC.GameStack.SavedEsi
        mov     edi,CC.GameStack.SavedEdi
        jmp     @@2

        { Back to the App stack }
      @@1:
        mov     CC.GameStack.SPtr,esp
        mov     CC.GameStack.SavedEbx,ebx
        mov     CC.GameStack.SavedEsi,esi
        mov     CC.GameStack.SavedEdi,edi
        mov     esp,CC.AppStack.SPtr     { Now on the new stack }
        mov     ebx,CC.AppStack.SavedEbx
        mov     esi,CC.AppStack.SavedEsi
        mov     edi,CC.AppStack.SavedEdi
      @@2:
        mov     ebp,esp
end;



procedure Spawn( Proc: TSpawnProc); assembler;
asm
        mov     eax,esp
        and     eax,0FFFF0000h
        { only an App can Spawn a game }
        cmp     eax,CC.AppStack.SBtm
        jne     @@1

        { Setup task stack }
        push    edi
        mov     edi,CC.GameStack.TTaskInfo.SPtr
        std

        {  Returning out will terminate task }
        mov     eax,OFFSET TerminateTask
        stosd

        { Setup ret to "call" proc }
        mov     eax,Proc
        stosd
        mov     eax,edi   { Fake EBP }
        stosd
        cld
        mov     CC.GameStack.TTaskInfo.SPtr,eax
        pop     edi

        push    0
        Call    Message
      @@1:
end;

end.

