{
          HELLO WORLD PROGRAM FOR WIN32
          Written in Virtual Pascal 2.1
          By L D Blake, Sept 2003

Program uses the VP Windows unit to demonstrate the basics of
window classes, window creation and message dispatching.

-------------------------------------------------------------------------
}

program HelloWorld;
{&PmType PM}
{&Delphi+,H+}

uses
  Windows;

var
  inst    : hwnd;
  hwind,
  hwind1,
  hwind2  : hwnd;
  WClass  : Twndclass;
  wmsg    : tmsg;

// Window message handler
Function WinProc(hwin : hwnd;   // window handle
                 wmsg : dword;  // window message
                 wparm: wparam; // first parameter
                 lparm: lparam  // second parameter
                 ) : longint; stdcall; export;
Begin
  Result := 0;
  Case wmsg of

    WM_Command :                    // returned by controls and menus
      case loword(wparm) of
        1 : begin
              postquitmessage(0);
            end;
      end;

    WM_destroy :                    // sent by the X on the border
      postquitmessage(0);

  Else                              // pass on unprocessed messages
    Result  := defwindowproc(hwin,wmsg,wparm,lparm);
  End;
End;

// Setup and create the window
procedure Setup;
begin
  inst := system.hinstance;             // VP provides program handle

  // First we need to register a window class
  With WClass do
    Begin
      style := cs_owndc;                  // put our class in it's own space
      lpfnWndProc   := @winproc;          // point to the message handler
      cbclsextra    := 0;                 // don't need extra class space
      cbwndextra    := 0;                 // no extra window space needed
      hinstance     := inst;              // program instance handle
      hicon         := 0;                 // use default window icon
      hcursor       := 0;                 // use default mouse cursor
      hbrbackground := 1 + color_window;  // use default white background
      lpszMenuName  := nil;               // no predefined menus
      lpszclassname :='Hello_Class'       // name of our class
    End;
    if registerclass(WClass) = 0 then     // register and test class
      begin
        messagebox(0,'Class registration failed','Oh-Oh',
                  mb_iconstop or mb_ok);
        halt(0);
      end;


  // Next We need a window
  hwind := createwindowex(ws_ex_clientedge, // sunken border
                  'Hello_Class',            // window Class
                  'Virtual Pascal',         // Window Title
                  ws_overlapped or          // make a border
                  ws_dlgframe or             // window has title bar
                  ws_minimizebox or         // window minimizes
                  ws_sysmenu or             // window has close button
                  ws_visible,               // and we want to see it
                  cw_usedefault,            // system sets left pos
                  cw_usedefault,            // system sets top
                  300,                      // set width
                  110,                      // set height
                  0,                        // no parent
                  0,                        // no menu return
                  inst,                     // instance handle
                  nil);                     // no extra data

  if hwind = 0 then                         // did the window create?
    begin
      messagebox(0,'Unable To Create Main Window','Oh-Oh',
                mb_iconstop or mb_ok);
      halt(0);
    end;

  // now we add a couple of controls
  hwind1 := createwindowex(ws_ex_clientedge,  // fancy sunken edge
                  'static',                   // text display panel
                  'Hello World! From Virtual Pascal', // display this
                  ws_child or                 // this is a control window
                  ws_dlgframe or              // fancy raised border
                  ws_visible or               // make it show up
                  ss_center,                  // center our text
                  20,                         // left relative to main window
                  10,                         // top relative to main window
                  250,                        // width of window
                  25,                         // height of window
                  hwind,                      // parented by main window
                  0,                          // no menu return
                  inst,                       // program handle
                  nil);                       // no extra data

   hwind2 := createwindow('button',           // this one's a button
                  'Hello Right Back At Ya',   // a little humour
                  ws_child or                 // another control window
                  ws_visible or               // we want to see it
                  bs_pushbutton,              // Square button type
                  20,                         // left edge
                  50,                         // top edge
                  250,                        // width
                  20,                         // height
                  hwind,                      // parent window handle
                  1,                          // send 1 to winproc
                  inst,                       // program handle
                  nil);                       // no extras

  // next we put it on screen
  showwindow(hwind,sw_shownormal);
  updatewindow(hwind);
end;

// Main program begins here
begin
  Setup;

  // A GUI application basically consists of a message dispatcher
  While getmessage(wmsg,0,0,0) do             // loop until wm_quit
    begin
      translatemessage(wmsg);
      dispatchmessage(wmsg);
    end;

  // And we're out of here
  halt(wmsg.wparam);
end.
