{}
{                                                       }
{      Virtual Pascal for Linux                         }
{      Resource support layer                           }
{      }
{      Copyright (C) 2000 vpascal.com                   }
{      Copyright (C) 1999 Veit Kannegieser              }
{                                                       }
{}

{$IFNDEF Linux}
  Compile/Target Platform: Linux
{$ENDIF}
{$S-}

unit LnxRes;

interface

type
  TLinux_Resource_Anchor = packed record
    fPointer   : Pointer;
    fLength    : Longint;
    fSignature : Array[0..3] of Char;
  end;

  // PE2LE: RESKONV.PAS: dpmi32_resource_block_type
  TLinux_Resource_Block = packed record
    fLength    : Longint;
    fType      : SmallWord;
    fId        : SmallWord;
  end;

  PLinux_Resource_String = ^TLinux_Resource_String;
  TLinux_Resource_String = packed record
    fStrLength : SmallWord;
    fStr       : Array[0..High(SmallWord)] of Char;
  end;

const
  Linux_Resource_Anchor : TLinux_Resource_Anchor = (
    fPointer   : Ptr($b582cc18);
    fLength    : $467a1d0e;
    fSignature : 'RES?' );

  rt_ResourceString     = 0;
  rt_RC_Data            = 1;
  rt_NameString         = 2;


function LnxGetResourceAddress(_Name, _Type: SmallWord): Pointer;
function LnxGetResourceStringAddress(_Name: SmallWord) : PLinux_Resource_String;
function LnxGetResourceAddressNamed(_Name: PChar; _Type: SmallWord;var _Size: Longint): Pointer;

implementation

Uses
  Strings;


function LnxGetResourceAddress(_Name, _Type: SmallWord): Pointer;
  assembler; {$Frame-} {&Uses esi,edi}
asm
      sub   eax,eax // not found
      cmp   [Linux_Resource_Anchor.fSignature],'OSER' // 'RESO'
      jne   @Ret

      mov   esi,[Linux_Resource_Anchor.fPointer]
      mov   edi,[Linux_Resource_Anchor.fLength]
      add   edi,esi

      mov   dx,[_Name]
      shl   edx,16
      mov   dx,[_Type]

      jmp   @BeginSearch

    @ContinueSearch:
      add   esi,[esi+TLinux_Resource_Block.fLength]

    @BeginSearch:
      // End of Resource ?
      cmp   esi,edi
      je    @Ret

      // Longint containing both type and id
      cmp   edx,[esi+TLinux_Resource_Block.fType].Longint
      jne   @ContinueSearch

      // Found: Return address
      lea   eax,[esi+type TLinux_Resource_Block]

    @Ret:
end;

function LnxGetResourceStringAddress(_Name: SmallWord) : PLinux_Resource_String;
  assembler; {&Frame-} {&Uses None}
asm
      movzx eax,[_Name]
      shr   eax,4   // div 16
      push  eax
      push  rt_ResourceString
      call  LnxGetResourceAddress

      test  eax,eax
      jz    @Ret

      movzx ecx,[_Name]
      and   ecx,15  // mod 16
      test  ecx,ecx
      jmp   @@1

    @SearchLoop:

      movzx edx,[eax].SmallWord
      lea   eax,[eax+2+edx]
      sub   ecx,1
    @@1:
      jnz   @SearchLoop

      cmp [eax].SmallWord,0
      jne @ret

      // string with length 0 -> nil
      sub eax,eax

    @ret:
end;

function LnxGetResourceAddressNamed(_Name: PChar; _Type: SmallWord;var _Size: Longint): Pointer;
  var
    work,next,sentinel  :^TLinux_Resource_Block;
    id_name_len         :Word;
  begin
    id_name_len:=StrLen(_Name);
    Result:=nil;
    with Linux_Resource_Anchor do
      if (Assigned(fPointer)) and (StrLComp(fSignature,'RESO',Length('RESO'))=0) then
        begin
          work:=fPointer;
          sentinel:=Ptr(Ofs(work)+fLength);

          // try numeric name id search
          if Longint(_Name)<=$ffff then
            while work<>sentinel do
              begin
                with work^ do
                  // type compare
                  if (fType=_Type)
                  // name id compare
                  and (fId=Longint(_Name)) then
                    begin
                      // found.
                      Result:=Ptr(Ofs(work^)+SizeOf(TLinux_Resource_Block));
                      _Size:=work^.fLength-SizeOf(TLinux_Resource_Block);
                      Exit;
                    end;

                // scan next block
                work:=Ptr(Ofs(work^)+work^.fLength);
              end;

          work:=fPointer;
          // try string name id search
          while work<>sentinel do
            begin
              with work^ do
                if fType=rt_NameString then
                  // name id record found
                  with PLinux_Resource_String(Ptr(Ofs(work^)+SizeOf(TLinux_Resource_Block)))^ do
                    // compare name id
                    if (fStrLength=id_name_len) and (StrLIComp(_Name,fStr,fStrLength)=0) then
                      begin
                        // look at following record
                        next:=Ptr(Ofs(work^)+fLength);
                        // search type correct?
                        if next^.fType=_Type then
                          begin
                            // yes, found.
                            Result:=Ptr(Ofs(next^)+SizeOf(TLinux_Resource_Block));
                            _Size:=next^.fLength-SizeOf(TLinux_Resource_Block);
                            Exit;
                          end;
                      end;

              // scan next block
              work:=Ptr(Ofs(work^)+work^.fLength);
            end;
        end;
  end;

end.
