unit lnxtrap;

interface

implementation

{$IfDef Linux}
uses
  Linux,
  VpSysLow,
  VpUtils;

type
  PFpReg = ^TFpReg;
  TFpReg = record
    losig:   LongInt;
    hisig:   LongInt;
    signexp: SmallWord;
  end;

  PContextRecord = ^TContextRecord;
  TContextRecord = record
    ContextFlags: LongInt;
    ctx_env:    array [0..6] of LongInt;
    ctx_stack:  array [0..7] of TFpReg;
    ctx_SegGs:  LongInt;
    ctx_SegFs:  LongInt;
    ctx_SegEs:  LongInt;
    ctx_SegDs:  LongInt;
    ctx_RegEdi: LongInt;
    ctx_RegEsi: LongInt;
    ctx_RegEax: LongInt;
    ctx_RegEbx: LongInt;
    ctx_RegEcx: LongInt;
    ctx_RegEdx: LongInt;
    ctx_RegEbp: LongInt;
    ctx_RegEip: LongInt;
    ctx_SegCs:  LongInt;
    ctx_EFlags: LongInt;
    ctx_RegEsp: LongInt;
    ctx_SegSs:  LongInt;
  end;

(*&Saves All*)
procedure show_exception(ArgCount,Arg1,Arg2,Code: Longint);
  const
    signame:array[SIGHUP..SIGUNUSED] of string[14]=

      ('SIGHUP',
       'SIGINT',
       'SIGQUIT',
       'SIGILL',
       'SIGTRAP',
       'SIGABRT/SIGIOT',
       'SIGBUS',
       'SIGFPE',
       'SIGKILL',
       'SIGUSR1',
       'SIGSEGV',
       'SIGUSR2',
       'SIGPIPE',
       'SIGALRM',
       'SIGTERM',
       'SIGSTKFLT',
       'SIGCHLD',
       'SIGCONT',
       'SIGSTOP',
       'SIGTSTP',
       'SIGTTIN',
       'SIGTTOU',
       'SIGURG',
       'SIGXCPU',
       'SIGXFSZ',
       'SIGVTALRM',
       'SIGPROF',
       'SIGWINCH',
       'SIGIO/SIGPOLL',
       'SIGPWR',
       'SIGUNUSED');

  var
    s:string;
    z:longint;
    loc:pointer;
    LineNo:longint;
  begin
    s:='Exception $'+Int2Hex(Code,8);
    case code of
      $0EEDFACF:s:=s+' (cLanguageReRaise)';
      $0EEDFAD0:s:=s+' (cLanguageExcept)';
      $0EEDFAD1:s:=s+' (cLanguageFinally)';
      $0EEDFAD2:s:=s+' (cLanguageTerminate)';
      $0EEDFAD3:s:=s+' (cLanguageUnhandled)';
      $0EEDFAD4:s:=s+' (cNonLanguageException)';
      $c0010000..
      $c01fffff:s:=s+' ('+signame[(code shr 16) and $ff]+')';
    end;
    LnxDebug(s);

    if (code>=$0EEDFACF) and (code<=$0EEDFAD3) then
      Exit;

    with PXcptReportRecord(Ptr(Arg1)) do
      begin
        s:='ExceptionNum=$'+Int2Hex(ExceptionNum,8)
          +' fHandlerFlags=$'+Int2Hex(fHandlerFlags,8);
        if NestedXcptReportRecord<>nil then
          s:=s+' NestedXcptReportRecord=$'+Int2Hex(Ofs(NestedXcptReportRecord^),8);
        LnxDebug(s);

        s:='ExceptionAddress=$'+Int2Hex(Ofs(ExceptionAddress^),8);
        for z:=1 to cParameters do
          s:=s+' ExceptionInfo['+Int2Str(z-1)+']=$'+Int2Hex(ExceptionInfo[z-1],8);
        LnxDebug(s);
      end;

    with PContextRecord(Ptr(Arg2)) do
      begin
        s:='CS=$'+Int2Hex(ctx_SegCs,8)+' EIP=$'+Int2Hex(ctx_RegEip,8);
        LnxDebug(s);
        s:='DS=$'+Int2Hex(ctx_SegDs,8)+' ESI=$'+Int2Hex(ctx_RegEsi,8);
        LnxDebug(s);
        s:='ES=$'+Int2Hex(ctx_SegEs,8)+' EDI=$'+Int2Hex(ctx_RegEdi,8);
        LnxDebug(s);
        s:='SS=$'+Int2Hex(ctx_SegSs,8)+' ESP=$'+Int2Hex(ctx_RegEsp,8)+' EBP=$'+Int2Hex(ctx_RegEbp,8);
        LnxDebug(s);
        s:='FS=$'+Int2Hex(ctx_SegFs,8);
        if ((ctx_SegFs and $7)=$7) then
          s:=s+' -> TID='+Int2Str(ctx_SegFs shr 3);
        LnxDebug(s);
        s:='GS=$'+Int2Hex(ctx_SegGs,8);
        LnxDebug(s);
        s:='EAX=$'+Int2Hex(ctx_RegEax,8)+' EBX=$'+Int2Hex(ctx_RegEbx,8)+' ECX=$'+Int2Hex(ctx_RegEcx,8)+' EDX=$'+Int2Hex(ctx_RegEdx,8);
        LnxDebug(s);
        s:='Flags=$'+Int2Hex(ctx_EFlags,8);
        LnxDebug(s);

        loc:=GetLocationInfo(Ptr(ctx_RegEip),s,LineNo);
        if Assigned(loc) then
          begin
            s:='[$'+Int2Hex(Ofs(loc^),8)+'='+s+'#'+Int2Str(LineNo)+']';
            LnxDebug(s);
          end;
      end;

  end;

(*&SAVES ebx,esi,edi*)

begin
  RaiseNotification_Hook:=@show_exception;
{$Else}
  (* for other platforms useable debuggers exist.. *)
{$EnDif}
end.

