{}
{                                                       }
{      Virtual Pascal for Linux                         }
{      PE to ELF conversion utility                     }
{      }
{      Copyright (C) 1999 Veit Kannegieser              }
{                                                       }
{}

(*$Use32+,I+,Delphi+,X+*)
(*  $D+,L+,Locinfo+*)
(*$M 1000000*)

program pe2elf;

// 1999.07.28 Veit Kannegieser
// ... VK/JP
// 1999.10.22 Version mit funkionierendem .so-Import an J.P. verschickt
// 2000.05.10 -v fr Sektionenliste und word->smallword fr PE2LETY.PAS
// 2000.12.17 realloc const32/4096, ld-linux.so.2, exe-beschreibung
// 2001.02.25 interpreter nur verwenden wenn importe vorhanden sind
//            -> sleep aus libc_6.so verwenden
// 2001.05.28 Resourcefehler behoben
// 2001.08.12 nicht benutzte Resourcen verwerfen
// 2001.12.09 FreeBSD brandelf-Funktionalitt eingebaut
// 2001.12.09 Unicode-Zeichensatz-Konverter mit und ohne Unit ZST (OnlyCP819)
// 2003.08.06 use pBufStream, use Unicode/Resource converter from pe2le
//            use some FillChar after ReallocMem to avaoid garbage in executable (ReallocMem does not zerofill!)

uses
  pe2lety,
  reskonv,
  Descript,
  ELFDef,
  dbwrite,
  Strings,
  Dos,
  Objects,
  VpUtils,
  VpSysLow;

const
  datum                         ='2003.08.06';
  memorymode                    :word={1}3;
  include_section_header        :boolean=true;
  resource_codepage             :string='ISO-8859-1';
  linebreak                     :array[1..2] of char=(^m,^j);
  descriptionframechar          :char='-';

type
  PCharArray                    = ^TCharArray;
  TCharArray                    = array[0..512*1024*1024      ] of Char;

  dynamic_tabelle_typ           =array[0..20000] of Elf32_Dyn_type;
  dynamic_tabelle_z_typ         =^dynamic_tabelle_typ;

  symboltabelle_typ             =array[0..200000] of Elf32_Sym_type;

  rel_plt_typ                   =array[0..200000] of Elf32_Rel_type;
  rel_plt_z_typ                 =^rel_plt_typ;

var
  d1,d2                         :pBufStream;
  pe_kopf                       :pe_kopf_typ;
  pe_obj_tab                    :array[pe_obj_typen] of pe_obj_typ;

  code_puffer,
  bss_puffer,
  data_puffer,
  import_puffer,
  relo_puffer,
  reso_temp_puffer,
  reso_puffer                   :pByteArray;

  plt_speicher                  :pByteArray;
  got_speicher                  :pLongArray;
  hash_tabelle                  :pLongArray;

  symboltabelle                 :^symboltabelle_typ;

  symbolstringtabelle           :PCharArray;

  rel_plt_speicher              :rel_plt_z_typ;

  importe_vorhanden,
  resourcen_vorhanden,
  bss_vorhanden                 :boolean;

  programm_kopf                 :array[0..40] of program_header_type;
  sektionen_kopf                :array[0..40] of section_header_type;
  speicher_tabelle              :array[0..40] of pointer;
  pk2                           :array[0..40] of program_header_type;

  string_puffer                 :pByteArray;
  string_puffer_laenge          :longint;
  reso_temp_laenge              :longint;
  anzahl_importe                :longint;

  InFile, OutFile               :string;

  programmlader                 :array[0..255] of char;

type
  sektionen=(sekt_null,         // reserviert
             sekt_interpreter,  // Programmlader
             sekt_text,         // Code
             sekt_bss,          // wenn vorhanden
             sekt_data_a,       // \ Konstanten, / alle initialisierten Seiten
             sekt_data_b,       // / Variablen   \ alle Nullseiten
             sekt_dynamic,      // Informationen/Zeiger fr den Programmlader
             sekt_got,          // Adressen importierter Funktionen
             sekt_dynsym,       // Tabelle importierte Symbole
             sekt_dynsymstr,    // Namen fr sekt_dynsym
             sekt_dynsymstrhash,// Hash-Tabelle fr sekt_dynsymstr
             sekt_rel_plt,      // Verbindung von plt zu dynsym
             sekt_resource,     // Resourcen
             sekt_plt,          // Aufrufe von importierten Funktionen
             sekt_strings);     // Namen von Sektionen


var
  sektionen_uebersicht:
    array[low(sektionen)..high(sektionen)] of
      record
        vorhanden               :boolean;
        index_                  :longint;
      end;

  anzahl_dynamic_eintraege      :longint;
  dynamic_speicher              :dynamic_tabelle_z_typ;


const
  sektionen_namen:array[low(sektionen)..high(sektionen)] {of array[1..2]} of string[25]=
    (('null,reserved'           ),
     ('interpreter'             ),
     ('.text (program code)'    ),
     ('"_bss"'                  ),
     ('data_a (.data)'          ),
     ('data_b (.bss)'           ),
     ('.dynamic'                ),
     ('.got'                    ),
     ('.dynsym'                 ),
     ('.dynsymstr'              ),
     ('.dynsymstrhash'          ),
     ('.rel.plt'                ),
     ('resource'                ),
     ('.plt'                    ),
     ('sektion header strings'  ));




  elf_kopf:Elf32_Ehdr_type=
   (magic0123                   :magic0123_elf;
    file_class                  :file_class__32bit;
    data_encoding               :ELFDATA2LSB;
    file_version                :e_version__current;
    osabi                       :ELFOSABI_NONE;
    abi_version                 :0;
    padding                     :(0,0,0,0,0,0,0);
    e_type                      :e_type__exec;
    e_machine                   :e_machine__intel386;
    e_version                   :e_version__current;
    e_entry                     :0;                     // entrypoint
    e_phoff                     :0;                     // program header offset
    e_shoff                     :0;                     // sections header offset
    e_flags                     :0;                     // keine
    e_ehsize                    :SizeOf(Elf32_Ehdr_type);
    e_phentsize                 :SizeOf(program_header_type);
    e_phnum                     :0;                     // 0..e_phnum-1 of entrys
    e_shentsize                 :SizeOf(section_header_type);
    e_shnum                     :0;                     // 0..e_shnum-1 of entrys
    e_shstrndx                  :0);

  delta=0;
  alignment=4;//1024; //{4096}1024;


procedure Abbruch(const zk:string);
  begin
    WriteLn(zk);
    Halt(255);
  end;

procedure Aufrunden(var a:longint;const m:longint);
  var
    rest:longint;
  begin
    rest:=a mod m;
    if rest<>0 then
      a:=a+m-rest;
  end;


//

procedure lade_pe;

  var
    pe_start            :longint;
    puffer              :array[0..$3f] of char;

  procedure lies(var p:pByteArray;var obj:pe_obj_typ);
    begin
      GetMem(p,obj.physical_size);
      d1^.Seek(obj.physical_offset);
      if d1^.Status<>stOK then
        abbruch('Stream seek error');
      d1^.Read(p^,obj.physical_size);
      if d1^.Status<>stOK then
        abbruch('Stream read error');
    end;

  procedure lies_obj_hdr(var h:pe_obj_typ);
    begin
      d1^.Read(h,SizeOf(h));
      if d1^.Status<>stOK then
        abbruch('Stream read error');

      (* kann nicht mehr aus Datei laden als Speicher reserviert ist *)
      if h.physical_size>h.virtual_size then
        h.physical_size:=h.virtual_size;
    end;

  begin
    {Assign(d1,);
    FileMode:=open_share_DenyNone+open_access_ReadOnly;
    Reset(d1,1);}
    d1:=New(pBufStream,Init(InFile,stOpenRead,512*1024));
    if d1^.Status<>stOK then
      abbruch('Can not open '+InFile+'!');

    pe_start:=0;
    repeat
      d1^.Read(puffer,SizeOf(puffer));
      if d1^.Status<>stOK then
        abbruch('Stream read error');

      if  (puffer[0]='P')
      and (puffer[1]='E')
      and (puffer[2]=#0 )
      and (puffer[3]=#0 )
       then
        Break;

      if ((puffer[0]='M') and (puffer[1]='Z'))
      or ((puffer[0]='Z') and (puffer[1]='M'))
       then
        begin
          pe_start:=pLongint(@puffer[$3c])^;
          d1^.Seek(pe_start);
          if d1^.Status<>stOK then
            abbruch('Stream seek error');
          Continue;
        end;


      if  (puffer[0]='L')
      and (puffer[1]='X')
      and (puffer[2]=#0 )
      and (puffer[3]=#0 )
       then
        abbruch('Error: source is LX');

      if  (puffer[0]='L')
      and (puffer[1]='E')
      and (puffer[2]=#0 )
      and (puffer[3]=#0 )
       then
        abbruch('Error: source is LE');

      if  (puffer[0]=#$7f)
      and (puffer[1]='E' )
      and (puffer[2]='L' )
      and (puffer[3]='F' )
       then
        abbruch('Error: source is ELF');

      abbruch('Error: unknown invalid source executable format.');

    until false;

    d1^.Seek(pe_start);
    if d1^.Status<>stOK then
      abbruch('Stream seek error');
    d1^.Read(pe_kopf,SizeOf(pe_kopf));
    if d1^.Status<>stOK then
      abbruch('Stream read error');

    (* obj-Tabelle laden

       expected object names:

       CODE32
       _BSS     (optional)
       CONST32
       .idata
       .edata   (optional)
       .reloc
       .rsrc    (optional)              *)

    if not (pe_kopf.num_obj in [4,5,6,7]) then
      abbruch('Error: not created with VP 2.x');

    d1^.Seek(pe_start+pe_kopf.nthdrsize+$18);
    if d1^.Status<>stOK then
      abbruch('Stream seek error');

    lies_obj_hdr(pe_obj_tab[code32 ]);
    lies_obj_hdr(pe_obj_tab[_bss   ]);
    bss_vorhanden:=(pe_obj_tab[_bss].name='_BSS'#0#0#0#0) or (pe_obj_tab[_bss].name='DATA32'#0#0);
    if not bss_vorhanden then
      begin
        pe_obj_tab[const32]:=pe_obj_tab[_bss];
        fillchar(pe_obj_tab[_bss],SizeOf(pe_obj_tab[_bss]),0);
      end
    else
      lies_obj_hdr(pe_obj_tab[const32]);
    lies_obj_hdr(pe_obj_tab[idata  ]);
    lies_obj_hdr(pe_obj_tab[reloc  ]);
    if pe_obj_tab[reloc  ].name<>'.reloc'#0#0 then (* .edata ? *)
      lies_obj_hdr(pe_obj_tab[reloc  ]);
    lies_obj_hdr(pe_obj_tab[res  ]);

    resourcen_vorhanden:=(pe_obj_tab[res    ].name='.rsrc'#0#0#0);
    if not resourcen_vorhanden then
      fillchar(pe_obj_tab[res  ],sizeof(pe_obj_typ),0);

    if  (pe_obj_tab[code32 ].name<>'CODE32'#0#0  )
    or ((pe_obj_tab[_bss   ].name<>'_BSS'#0#0#0#0) and (pe_obj_tab[_bss   ].name<>'DATA32'#0#0) and bss_vorhanden)
    or  (pe_obj_tab[const32].name<>'CONST32'#0   )
    or  (pe_obj_tab[idata  ].name<>'.idata'#0#0  )
    or  (pe_obj_tab[reloc  ].name<>'.reloc'#0#0  )
    or ((pe_obj_tab[res    ].name<>'.rsrc'#0#0#0 ) and resourcen_vorhanden)
     then
      abbruch('Error: not created with VP 2.x');

    importe_vorhanden:={}(pe_obj_tab[idata  ].physical_size>$14){}{true};

    lies(code_puffer,pe_obj_tab[code32 ]);
    if bss_vorhanden then
      lies(bss_puffer,pe_obj_tab[_bss]);
    lies(data_puffer,pe_obj_tab[const32]);
    lies(import_puffer,pe_obj_tab[idata]);
    lies(relo_puffer,pe_obj_tab[reloc]);
    if resourcen_vorhanden then
      lies(reso_puffer,pe_obj_tab[res]);

    d1^.Done;

    load_module_description(InFile);
  end;

//


procedure rechne;
  var
    dateiposition,adressposition:longint;
    sekt_z:sektionen;
    index_:longint;
    z1:longint;

  function name_speichern(const zk:string):longint;
    begin
      Result:=string_puffer_laenge;
      ReallocMem(string_puffer,string_puffer_laenge+Length(zk)+1);
      Move(zk[1],string_puffer^[string_puffer_laenge],Length(zk));
      Inc(string_puffer_laenge,Length(zk));
      string_puffer^[string_puffer_laenge]:=0; // #0
      Inc(string_puffer_laenge);
    end;

  procedure suche_naechste_passende_dateipostion;
    begin
      // mod 4096 = and $fff
      if (dateiposition and $fff)<>(adressposition and $fff) then
        begin
          dateiposition:=(dateiposition and $fffff000)+$1000+(adressposition and $fff);
        end;
    end;

  procedure suche_naechste_passende_adressposition;
    begin
      // mod 4096 = and $fff
      if (dateiposition and $fff)<>(adressposition and $fff) then
        begin
          adressposition:=(adressposition and $fffff000)+$1000+(dateiposition and $fff);
        end;
    end;

  (*$I impconve.pas*)
  (*$I proghead.pas*)
  (*$I search_d.pas*)

  var
    oldcodesize         :Longint;
    olddatasize         :Longint;
  begin
    search_description;

    string_puffer_laenge:=0;
    string_puffer:=nil;
    name_speichern('');

    anzahl_dynamic_eintraege:=0;
    dynamic_speicher:=nil;

    // Lnge eines Symboltabelleneintrages
    Speichere_in_der_DYNAMIC_Tabelle(DT_SYMENT,SizeOf(Elf32_Sym_type));

    // Code-Bereich aufblhen ? - bringt nichts
    if memorymode>=3+1 then
      begin
        oldcodesize:=pe_obj_tab[code32].physical_size;
        Aufrunden(pe_obj_tab[code32].virtual_size ,4096);
        Aufrunden(pe_obj_tab[code32].physical_size,4096);
        ReallocMem(code_puffer,pe_obj_tab[code32].physical_size);

        FillChar(code_puffer^[oldcodesize],pe_obj_tab[code32].physical_size-oldcodesize,0);
      end;


    // das von VP im Datenbereich angelegte Stacksegment kann gelscht werden
    z1:=Meml[Ofs(code_puffer^[pe_kopf.entrypoint-pe_obj_tab[code32].rva])+5+1];
    Dec(pe_obj_tab[const32].virtual_size,z1);

    // Nullen am Ende des Datenbereiches kommen in den virtuellen Teil
    with pe_obj_tab[const32] do
      begin
        olddatasize:=physical_size;

        while data_puffer^[physical_size-1-1]=0 do
          Dec(physical_size);
        // auf 32-Bit-Lnge bringen
        Aufrunden(physical_size,4);

        if memorymode>=3 then // noch schlimmer aufblhen
          begin
            if virtual_size<physical_size then
              virtual_size:=physical_size;
            Aufrunden(virtual_size ,4096);
            physical_size:=virtual_size;
          end
        else if memorymode>=2 then
          begin
            Aufrunden(physical_size,4096);
            Aufrunden(virtual_size,4096);
            Inc(virtual_size,4096*10);
          end;
      end;

    ReAllocMem(data_puffer,pe_obj_tab[const32].physical_size);
    if pe_obj_tab[const32].physical_size>olddatasize then
      FillChar(data_puffer^[olddatasize],pe_obj_tab[const32].physical_size-olddatasize,0);

    // welche Sektionen werden bentigt ?
    FillChar(sektionen_uebersicht,SizeOf(sektionen_uebersicht),0); // 0,false,..
    sektionen_uebersicht[sekt_null].vorhanden:=true;
    sektionen_uebersicht[sekt_text].vorhanden:=true;    (* Code *)
    sektionen_uebersicht[sekt_data_a].vorhanden:=true;  (* Const *)
    if pe_obj_tab[const32].virtual_size>pe_obj_tab[const32].physical_size then
      sektionen_uebersicht[sekt_data_b].vorhanden:=true; (* Variables *)

    if importe_vorhanden then
      begin
        sektionen_uebersicht[sekt_interpreter   ].vorhanden:=true;
        sektionen_uebersicht[sekt_plt           ].vorhanden:=true;
        sektionen_uebersicht[sekt_rel_plt       ].vorhanden:=true;
        sektionen_uebersicht[sekt_got           ].vorhanden:=true;
        sektionen_uebersicht[sekt_dynamic       ].vorhanden:=true;
        sektionen_uebersicht[sekt_dynsym        ].vorhanden:=true;
        sektionen_uebersicht[sekt_dynsymstr     ].vorhanden:=true;
        sektionen_uebersicht[sekt_dynsymstrhash ].vorhanden:=true;
      end;

    // Resourceumwandlung
    if resourcen_vorhanden then
      (* if you stop here with an compiler error, then place Use_PE2Elf
         into the Options/Compiler/conditional defines field and recompile all *)
      if not search_resource_anchor(data_puffer,pe_obj_tab[const32].physical_size) then
        begin
          WriteLn('Dropping ',pe_obj_tab[res].physical_size,' bytes unused Resources.');
          resourcen_vorhanden:=false;
          Dispose(reso_puffer);
          reso_puffer:=nil;
          FillChar(pe_obj_tab[res],SizeOf(pe_obj_tab[res]),0);
        end;


    if resourcen_vorhanden then
      begin
        // sekt_resource
        sektionen_uebersicht[sekt_resource].vorhanden:=true;
        (* Annahme: viel weniger Platz wird bentigt *)
        GetMem(reso_temp_puffer,(pe_obj_tab[res].physical_size*3) div 2);
        reso_temp_laenge:=0;

        konvertiere_resourcen(reso_puffer,pe_obj_tab[res].physical_size,pe_obj_tab[res].rva,
                              nil,0,
                              reso_temp_puffer,reso_temp_laenge,resource_codepage);

        (* neu in richtiger Gre anfordern *)
        ReAllocMem(reso_puffer,reso_temp_laenge);
        (* Daten Kopieren *)
        Move(reso_temp_puffer^,reso_puffer^,reso_temp_laenge);
        (* temporren Speicher freigeben *)
        Dispose(reso_temp_puffer);
        (* neue Lnge merken *)

        pe_obj_tab[res].virtual_size :=reso_temp_laenge;
        pe_obj_tab[res].physical_size:=reso_temp_laenge;
      end;

    if include_section_header then
      sektionen_uebersicht[sekt_strings].vorhanden:=true;

    // Nummern verteilen und zhlen
    for sekt_z:=low(sektionen) to high(sektionen) do
      if sektionen_uebersicht[sekt_z].vorhanden then
        begin
          sektionen_uebersicht[sekt_z].index_:=elf_kopf.e_shnum;
          Inc(elf_kopf.e_shnum);
          Inc(elf_kopf.e_phnum);
        end
      else
        sektionen_uebersicht[sekt_z].index_:=-1;

    if not include_section_header then
      elf_kopf.e_shnum:=0;


    //*********************************************************************

    // Sektionen fllen und Dateipostionen vergeben
    dateiposition:=SizeOf(elf_kopf);
    if module_description<>'' then
      begin
        (* ^m^j
            -----^m^j
            descr^m^j
            -----^m^j *)
        Inc(dateiposition,4*2+3*Length(module_description));
      end;
    if include_section_header then
      begin
        elf_kopf.e_shoff:=dateiposition;
        Inc(dateiposition,elf_kopf.e_shentsize*elf_kopf.e_shnum);
      end;

    FillChar(sektionen_kopf[0],SizeOf(sektionen_kopf),0); // SHT_NULL
    FillChar(programm_kopf [0],SizeOf(programm_kopf ),0); // PT_NULL

    adressposition:=$00401000-$1000+dateiposition;
    //***********************************************************************
    // Programmlader
    if sektionen_uebersicht[sekt_interpreter].vorhanden then
      begin
        index_:=sektionen_uebersicht[sekt_interpreter].index_;
        with sektionen_kopf[index_] do
          begin
            sh_name         :=name_speichern('.interp');
            sh_type         :=SHT_PROGBITS;
            sh_flags        :=SHF_ALLOC;
            sh_addr         :=adressposition;
            sh_offset       :=dateiposition;
            sh_size         :=StrLen(programmlader)+1;
            sh_link         :=0;
            sh_info         :=0;
            sh_addralign    :=1;
            sh_entsize      :=0;
          end;
        with programm_kopf [index_] do
          begin
            p_type          :=PT_INTERP;
            p_offset        :=sektionen_kopf[index_].sh_offset;
            p_vaddr         :=sektionen_kopf[index_].sh_addr;
            p_paddr         :=0;
            p_filesz        :=sektionen_kopf[index_].sh_size;
            p_memsz         :=sektionen_kopf[index_].sh_size;
            p_flags         :=PF_R;
            p_align         :=sektionen_kopf[index_].sh_addralign;
          end;

        Inc(adressposition,sektionen_kopf[index_].sh_size);
        Inc(dateiposition ,sektionen_kopf[index_].sh_size);

        Aufrunden(adressposition,4);
        Aufrunden(dateiposition ,4);
      end;

    //***********************************************************************
    // CODE bearbeiten
    // $00401000 damit der VP-debugger funktioniert
    adressposition:=pe_obj_tab[code32].rva+pe_kopf.image_base+delta;
    suche_naechste_passende_dateipostion;

    index_:=sektionen_uebersicht[sekt_text].index_;
    with sektionen_kopf[index_] do
      begin
        sh_name         :=name_speichern('.text');
        sh_type         :=SHT_PROGBITS;
        sh_flags        :=SHF_ALLOC+SHF_EXECINSTR;
        sh_addr         :=adressposition;
        sh_offset       :=dateiposition;
        sh_size         :=pe_obj_tab[code32].physical_size;
        sh_link         :=0;
        sh_info         :=0;
        sh_addralign    :=4;
        sh_entsize      :=0;
      end;
    with programm_kopf [index_] do
      begin
        p_type          :=PT_LOAD;
        p_offset        :=sektionen_kopf[index_].sh_offset;
        p_vaddr         :=sektionen_kopf[index_].sh_addr;
        p_paddr         :=0;
        p_filesz        :=sektionen_kopf[index_].sh_size;
        p_memsz         :=sektionen_kopf[index_].sh_size;
        p_flags         :=PF_X+PF_R;
        p_align         :=sektionen_kopf[index_].sh_addralign;
      end;

    // Eintrittspunkt
    elf_kopf.e_entry:=pe_kopf.entrypoint+pe_kopf.image_base+delta;

    Inc(adressposition,sektionen_kopf[index_].sh_size);
    Inc(dateiposition ,sektionen_kopf[index_].sh_size);

    Aufrunden(adressposition,4);
    Aufrunden(dateiposition ,4);

    //***********************************************************************
    // BSS
    if sektionen_uebersicht[sekt_bss].vorhanden then
      begin
        adressposition:=pe_obj_tab[_bss].rva+pe_kopf.image_base+delta;

        index_:=sektionen_uebersicht[sekt_bss].index_;
        with sektionen_kopf[index_] do
          begin
            sh_name         :=name_speichern('.bss');
            sh_type         :=SHT_NOBITS;
            sh_flags        :=SHF_ALLOC+SHF_WRITE;
            sh_addr         :=adressposition;
            sh_offset       :=0;
            sh_size         :=pe_obj_tab[_bss].virtual_size;
            sh_link         :=0;
            sh_info         :=0;
            sh_addralign    :=4;
            sh_entsize      :=0;
          end;
        with programm_kopf [index_] do
          begin
            p_type          :=PT_LOAD;
            p_offset        :=sektionen_kopf[index_].sh_offset;
            p_vaddr         :=sektionen_kopf[index_].sh_addr;
            p_paddr         :=0;
            p_filesz        :=0;
            p_memsz         :=sektionen_kopf[index_].sh_size;
            p_flags         :=PF_R+PF_W;
            p_align         :=sektionen_kopf[index_].sh_addralign;
          end;

        Inc(adressposition,sektionen_kopf[index_].sh_size);
        Aufrunden(adressposition,4);
      end;

    //***********************************************************************
    // Konstanten und Daten - Teil 1
    adressposition:=pe_obj_tab[const32].rva+pe_kopf.image_base+delta;
    suche_naechste_passende_dateipostion;

    index_:=sektionen_uebersicht[sekt_data_a].index_;
    with sektionen_kopf[index_] do
      begin
        sh_name         :=name_speichern('.data');
        sh_type         :=SHT_PROGBITS;
        sh_flags        :=SHF_ALLOC+SHF_WRITE;
        sh_addr         :=adressposition;
        sh_offset       :=dateiposition;
        sh_size         :=pe_obj_tab[const32].physical_size;
        sh_link         :=0;
        sh_info         :=0;
        sh_addralign    :=4;
        sh_entsize      :=0;
      end;
    with programm_kopf [index_] do
      begin
        p_type          :=PT_LOAD;
        p_offset        :=sektionen_kopf[index_].sh_offset;
        p_vaddr         :=sektionen_kopf[index_].sh_addr;
        p_paddr         :=0;
        p_filesz        :=sektionen_kopf[index_].sh_size;
        p_memsz         :=sektionen_kopf[index_].sh_size;
        p_flags         :=PF_R+PF_W;
        p_align         :=sektionen_kopf[index_].sh_addralign;
      end;

    Inc(adressposition,sektionen_kopf[index_].sh_size);
    Inc(dateiposition ,sektionen_kopf[index_].sh_size);

    //***********************************************************************
    // Konstanten und Daten - Teil 2
    if sektionen_uebersicht[sekt_data_b].vorhanden then
      begin
        index_:=sektionen_uebersicht[sekt_data_b].index_;
        with sektionen_kopf[index_] do
          begin
            sh_name         :=name_speichern('.bss');
            sh_type         :=SHT_NOBITS;
            sh_flags        :=SHF_ALLOC+SHF_WRITE;
            sh_addr         :=adressposition;
            sh_offset       :=dateiposition;
            sh_size         :=pe_obj_tab[const32].virtual_size-pe_obj_tab[const32].physical_size;
            sh_link         :=0;
            sh_info         :=0;
            sh_addralign    :=4;
            sh_entsize      :=0;
          end;
        with programm_kopf [index_] do
          begin
            p_type          :=PT_LOAD;
            p_offset        :=sektionen_kopf[index_].sh_offset;
            p_vaddr         :=sektionen_kopf[index_].sh_addr;
            p_paddr         :=0;
            p_filesz        :=0;
            p_memsz         :=sektionen_kopf[index_].sh_size;
            p_flags         :=PF_R+PF_W;
            p_align         :=sektionen_kopf[index_].sh_addralign;
          end;
        Inc(adressposition,sektionen_kopf[index_].sh_size);
        Aufrunden(adressposition,4);
        suche_naechste_passende_adressposition;
      end;

    resource_adressposition:=0;
    if importe_vorhanden then
      import_umrechnung
    else
      begin
        resource_laenge:=pe_obj_tab[res].physical_size;
        resource_adressposition:=adressposition;
        resource_dateiposition :=dateiposition ;
        Inc(adressposition,resource_laenge);
        Inc(dateiposition ,resource_laenge);
      end;


    //***********************************************************************
    // ".dynamic"
    if sektionen_uebersicht[sekt_dynamic].vorhanden then
      begin
        index_:=sektionen_uebersicht[sekt_dynamic].index_;
        with sektionen_kopf[index_] do
          begin
            sh_name         :=name_speichern('.dynamic');
            sh_type         :=SHT_DYNAMIC;
            sh_flags        :=SHF_WRITE+SHF_ALLOC;
            sh_addr         :=_dynamic_adressposition;
            sh_offset       :=_dynamic_dateiposition ;
            sh_size         :=_dynamic_laenge;
            sh_link         :=sektionen_uebersicht[sekt_dynsymstr].index_;
            sh_info         :=0;
            sh_addralign    :=4;
            sh_entsize      :=SizeOf(Elf32_Dyn_type);
          end;
        with programm_kopf [index_] do
          begin
            p_type          :=PT_DYNAMIC;
            p_offset        :=sektionen_kopf[index_].sh_offset;
            p_vaddr         :=sektionen_kopf[index_].sh_addr;
            p_paddr         :=0;
            p_filesz        :=sektionen_kopf[index_].sh_size;
            p_memsz         :=sektionen_kopf[index_].sh_size;
            p_flags         :=PF_R+PF_W;
            p_align         :=sektionen_kopf[index_].sh_addralign;
          end;
      end;

    //***********************************************************************
    // ".got"
    if sektionen_uebersicht[sekt_got].vorhanden then
      begin
        index_:=sektionen_uebersicht[sekt_got].index_;
        with sektionen_kopf[index_] do
          begin
            sh_name         :=name_speichern('.got');
            sh_type         :=SHT_PROGBITS;
            sh_flags        :=SHF_WRITE+SHF_ALLOC;
            sh_addr         :=_got_adressposition;
            sh_offset       :=_got_dateiposition ;
            sh_size         :=_got_laenge;
            sh_link         :=0;
            sh_info         :=0;
            sh_addralign    :=4;
            sh_entsize      :=4;
          end;
        with programm_kopf [index_] do
          begin
            p_type          :=PT_LOAD;
            p_offset        :=sektionen_kopf[index_].sh_offset;
            p_vaddr         :=sektionen_kopf[index_].sh_addr;
            p_paddr         :=0;
            p_filesz        :=sektionen_kopf[index_].sh_size;
            p_memsz         :=sektionen_kopf[index_].sh_size;
            p_flags         :=PF_R+PF_W;
            p_align         :=sektionen_kopf[index_].sh_addralign;
          end;
      end;


    //***********************************************************************
    // ".dynsym"
    if sektionen_uebersicht[sekt_dynsym].vorhanden then
      begin
        index_:=sektionen_uebersicht[sekt_dynsym].index_;
        with sektionen_kopf[index_] do
          begin
            sh_name         :=name_speichern('.dynsym');
            sh_type         :=SHT_DYNSYM;
            sh_flags        :=SHF_ALLOC;
            sh_addr         :=_dynsym_adressposition;
            sh_offset       :=_dynsym_dateiposition ;
            sh_size         :=_dynsym_laenge;
            sh_link         :=sektionen_uebersicht[sekt_dynsymstr].index_;
            sh_info         :=sektionen_uebersicht[sekt_interpreter].index_;
            sh_addralign    :=4;
            sh_entsize      :=SizeOf(Elf32_Sym_type);
          end;
        with programm_kopf [index_] do
          begin
            p_type          :=PT_LOAD;
            p_offset        :=sektionen_kopf[index_].sh_offset;
            p_vaddr         :=sektionen_kopf[index_].sh_addr;
            p_paddr         :=0;
            p_filesz        :=sektionen_kopf[index_].sh_size;
            p_memsz         :=sektionen_kopf[index_].sh_size;
            p_flags         :=PF_R;
            p_align         :=sektionen_kopf[index_].sh_addralign;
          end;
      end;

    //***********************************************************************
    // ".dynstr"
    if sektionen_uebersicht[sekt_dynsymstr].vorhanden then
      begin
        index_:=sektionen_uebersicht[sekt_dynsymstr].index_;
        with sektionen_kopf[index_] do
          begin
            sh_name         :=name_speichern('.dynstr');
            sh_type         :=SHT_STRTAB;
            sh_flags        :=SHF_ALLOC;
            sh_addr         :=_dynsymstr_adressposition;
            sh_offset       :=_dynsymstr_dateiposition;
            sh_size         :=symbolstringtabelle_laenge;
            sh_link         :=0;
            sh_info         :=0;
            sh_addralign    :=4;
            sh_entsize      :=0;
          end;
        with programm_kopf [index_] do
          begin
            p_type          :=PT_LOAD;
            p_offset        :=sektionen_kopf[index_].sh_offset;
            p_vaddr         :=sektionen_kopf[index_].sh_addr;
            p_paddr         :=0;
            p_filesz        :=sektionen_kopf[index_].sh_size;
            p_memsz         :=sektionen_kopf[index_].sh_size;
            p_flags         :=PF_R;
            p_align         :=sektionen_kopf[index_].sh_addralign;
          end;
      end;

    //***********************************************************************
    // ".hash"
    if sektionen_uebersicht[sekt_dynsymstrhash].vorhanden then
      begin
        index_:=sektionen_uebersicht[sekt_dynsymstrhash].index_;
        with sektionen_kopf[index_] do
          begin
            sh_name         :=name_speichern('.hash');
            sh_type         :=SHT_HASH;
            sh_flags        :=SHF_ALLOC;
            sh_addr         :=_dynsymstrhash_adressposition;
            sh_offset       :=_dynsymstrhash_dateiposition;
            sh_size         :=hash_tabelle_laenge;
            sh_link         :=sektionen_uebersicht[sekt_dynsym].index_;
            sh_info         :=0;
            sh_addralign    :=4;
            sh_entsize      :=4;
          end;
        with programm_kopf [index_] do
          begin
            p_type          :=PT_LOAD;
            p_offset        :=sektionen_kopf[index_].sh_offset;
            p_vaddr         :=sektionen_kopf[index_].sh_addr;
            p_paddr         :=0;
            p_filesz        :=sektionen_kopf[index_].sh_size;
            p_memsz         :=sektionen_kopf[index_].sh_size;
            p_flags         :=PF_R;
            p_align         :=sektionen_kopf[index_].sh_addralign;
          end;
      end;

    //***********************************************************************
    // Resourcen
    if sektionen_uebersicht[sekt_resource].vorhanden then
      begin
        index_:=sektionen_uebersicht[sekt_resource].index_;
        with sektionen_kopf[index_] do
          begin
            sh_name         :=name_speichern('resource');
            sh_type         :=SHT_PROGBITS;
            sh_flags        :=SHF_ALLOC;
            sh_addr         :=resource_adressposition;
            sh_offset       :=resource_dateiposition;
            sh_size         :=resource_laenge;
            sh_link         :=0;
            sh_info         :=0;
            sh_addralign    :=4;
            sh_entsize      :=0;
          end;
        with programm_kopf [index_] do
          begin
            p_type          :=PT_LOAD;
            p_offset        :=sektionen_kopf[index_].sh_offset;
            p_vaddr         :=sektionen_kopf[index_].sh_addr;
            p_paddr         :=0;
            p_filesz        :=sektionen_kopf[index_].sh_size;
            p_memsz         :=sektionen_kopf[index_].sh_size;
            p_flags         :=PF_R;
            p_align         :=4;
          end;
      end;
    //***********************************************************************
    // ".rel.plt"
    if sektionen_uebersicht[sekt_rel_plt].vorhanden then
      begin
        index_:=sektionen_uebersicht[sekt_rel_plt].index_;
        with sektionen_kopf[index_] do
          begin
            sh_name         :=name_speichern('.rel.plt');
            sh_type         :=SHT_REL;
            sh_flags        :=SHF_ALLOC;
            sh_addr         :=_rel_plt_adressposition;
            sh_offset       :=_rel_plt_dateiposition;
            sh_size         :=_rel_plt_laenge;
            sh_link         :=sektionen_uebersicht[sekt_dynsym].index_;
            sh_info         :=sektionen_uebersicht[sekt_plt   ].index_;
            sh_addralign    :=4;
            sh_entsize      :=SizeOf(Elf32_Rel_type);
          end;
        with programm_kopf [index_] do
          begin
            p_type          :=PT_LOAD;
            p_offset        :=sektionen_kopf[index_].sh_offset;
            p_vaddr         :=sektionen_kopf[index_].sh_addr;
            p_paddr         :=0;
            p_filesz        :=sektionen_kopf[index_].sh_size;
            p_memsz         :=sektionen_kopf[index_].sh_size;
            p_flags         :=PF_R;
            p_align         :=sektionen_kopf[index_].sh_addralign;
          end;
      end;


    //***********************************************************************
    // ".plt"
    if sektionen_uebersicht[sekt_plt].vorhanden then
      begin
        index_:=sektionen_uebersicht[sekt_plt].index_;
        with sektionen_kopf[index_] do
          begin
            sh_name         :=name_speichern('.plt');
            sh_type         :=SHT_PROGBITS;
            sh_flags        :=SHF_ALLOC+SHF_EXECINSTR;
            sh_addr         :=_plt_adressposition;
            sh_offset       :=_plt_dateiposition;
            sh_size         :=_plt_laenge;
            sh_link         :=0;
            sh_info         :=0;
            sh_addralign    :=4;
            sh_entsize      :=0;
          end;
        with programm_kopf [index_] do
          begin
            p_type          :=PT_LOAD;
            p_offset        :=sektionen_kopf[index_].sh_offset;
            p_vaddr         :=sektionen_kopf[index_].sh_addr;
            p_paddr         :=0;
            p_filesz        :=sektionen_kopf[index_].sh_size;
            p_memsz         :=sektionen_kopf[index_].sh_size;
            p_flags         :=PF_R+PF_X;
            p_align         :=sektionen_kopf[index_].sh_addralign;
          end;
      end;


    //***********************************************************************
    // Sektionsnamensektion
    if sektionen_uebersicht[sekt_strings].vorhanden then
      begin
        index_:=sektionen_uebersicht[sekt_strings].index_;
        with sektionen_kopf[index_] do
          begin
            sh_name         :=name_speichern('.shstrtab');
            sh_type         :=SHT_STRTAB;
            sh_flags        :=0;
            sh_addr         :=0;
            sh_offset       :=dateiposition;
            sh_size         :=string_puffer_laenge;
            sh_link         :=0;
            sh_info         :=0;
            sh_addralign    :=4;
            sh_entsize      :=0;
          end;
        with programm_kopf [index_] do
          begin
            p_type          :=PT_NULL;
            p_offset        :=sektionen_kopf[index_].sh_offset;
            p_vaddr         :=sektionen_kopf[index_].sh_addr;
            p_paddr         :=0;
            p_filesz        :=string_puffer_laenge;
            p_memsz         :=string_puffer_laenge;
            p_flags         :=0;
            p_align         :=sektionen_kopf[index_].sh_addralign;
          end;
        Inc(dateiposition ,string_puffer_laenge);
        Inc(adressposition,string_puffer_laenge);
        elf_kopf.e_shstrndx:=sektionen_uebersicht[sekt_strings].index_;
      end;

    //**********************************************************************
    // wenn Resourcen vorhanden sind, mu der Anker in Resource.pas
    // gefunden und angepat werden
    if sektionen_uebersicht[sekt_resource].vorhanden then
      begin
        patch_resource_anchor(data_puffer,
          sektionen_kopf[sektionen_uebersicht[sekt_resource].index_].sh_addr,
          sektionen_kopf[sektionen_uebersicht[sekt_resource].index_].sh_size);
      end;

    bearbeite_programmheader;

    {
    for z1:=1 to e_phnum do
      tausch[z1]:=z1;



    for z1:=1 to e_phnum-1 do
      for z2:=z1+1 to e_phnum do
        if programm_kopf [index_]}

  end;

//

procedure schreibe_elf;
  var
    d2_pos,
    d2_laenge           :longint;

  procedure BlockWrite_d2(const position:longint;var p;const laenge:longint);
    begin
      if d2_pos<>position then
        begin
          d2^.Seek(position);
          if d2^.Status<>stOK then
            abbruch('Stream seek error');
        end;
      d2^.Write(p,laenge);
      if d2^.Status<>stOK then
        abbruch('Stream write error');
      d2_pos:=position+laenge;
      d2_laenge:=Max(d2_laenge,d2_pos);
    end;

  procedure Fuellen(const anzahl:longint);
    var
      null_feld:array[0..4096-1] of byte;
    begin
      if anzahl<=0 then Exit;

      DWriteLn('***'+'  '+
               Int2Hex(d2_laenge,8)+'  '+
               Int2Hex(anzahl,8)
               {'********'+'  ',});

      FillChar(null_feld,anzahl,0);
      BlockWrite_d2(d2_laenge,null_feld,anzahl);
    end;

  var
    sekt_z              :sektionen;
    laenge              :longint;

  begin
    d2:=New(pBufStream,Init(OutFile,stCreate,512*1024));
    if d2^.Status<>stOK then
      abbruch('Error creating '+OutFile+'!');
    d2_pos:=0;
    d2_laenge:=0;

    BlockWrite_d2(d2_pos,elf_kopf,SizeOf(elf_kopf));

    if module_description<>'' then
      begin
        (* ^m^j
            -----^m^j
            descr^m^j
            -----^m^j *)
        BlockWrite_d2(d2_pos,linebreak,SizeOf(linebreak));
        for laenge:=1 to Length(module_description) do
          BlockWrite_d2(d2_pos,descriptionframechar,Sizeof(descriptionframechar));
        BlockWrite_d2(d2_pos,linebreak,SizeOf(linebreak));
        BlockWrite_d2(d2_pos,module_description[1],Length(module_description));
        BlockWrite_d2(d2_pos,linebreak,SizeOf(linebreak));
        for laenge:=1 to Length(module_description) do
          BlockWrite_d2(d2_pos,descriptionframechar,Sizeof(descriptionframechar));
        BlockWrite_d2(d2_pos,linebreak,SizeOf(linebreak));
      end;

    if include_section_header then
      BlockWrite_d2(d2_pos,sektionen_kopf,elf_kopf.e_shentsize*elf_kopf.e_shnum);

    DWriteLn('sec'+'  '+
             'filepos '+'  '+
             'len     '+'  '+
             'addr    '+'  '+
             'len     '+'  '+
             'attr  '+
            {'Name der Sektion'}'name of section');

    for sekt_z:=sekt_interpreter to high(sektionen) do
      if sektionen_uebersicht[sekt_z].vorhanden then
        begin
          laenge:=programm_kopf[sektionen_uebersicht[sekt_z].index_].p_filesz;

          if (laenge>0) and (sektionen_kopf[sektionen_uebersicht[sekt_z].index_].sh_offset>0) then
            Fuellen(sektionen_kopf[sektionen_uebersicht[sekt_z].index_].sh_offset-d2_laenge);

          DWrite  (Int2Hex(Ord(sektionen_uebersicht[sekt_z].index_),3)+'  '+
                   {Int2Hex(d2_pos,8),'  ',}
                   Int2Hex(sektionen_kopf[sektionen_uebersicht[sekt_z].index_].sh_offset,8)+'  '+
                   Int2Hex(laenge,8)+'  '+
                   Int2Hex(sektionen_kopf[sektionen_uebersicht[sekt_z].index_].sh_addr,8)+'  '+
                   Int2Hex(programm_kopf[sektionen_uebersicht[sekt_z].index_].p_memsz,8)+'  ');

          with programm_kopf[sektionen_uebersicht[sekt_z].index_] do
            begin
              if (p_flags and PF_R)=PF_R then
                DWrite('R')
              else
                DWrite('-');
              if (p_flags and PF_W)=PF_W then
                DWrite('W')
              else
                DWrite('-');
              if (p_flags and PF_X)=PF_X then
                DWrite('X')
              else
                DWrite('-');
              DWrite('   ');
            end;

          DWriteLn(sektionen_namen[sekt_z{,1}]);

          if laenge>0 then
            if d2_pos<>programm_kopf[sektionen_uebersicht[sekt_z].index_].p_offset then
              begin
                d2_pos:=programm_kopf[sektionen_uebersicht[sekt_z].index_].p_offset;
                d2^.Seek(d2_pos);
                if d2^.Status<>stOK then
                  abbruch('Stream seek error');
              end;

          if (laenge>0) and (programm_kopf[sektionen_uebersicht[sekt_z].index_].p_vaddr>0) then
            if (d2_pos mod 4096)<>(programm_kopf[sektionen_uebersicht[sekt_z].index_].p_vaddr mod 4096) then
              //if memorymode>=2 then//!!!!!
                begin
                  WriteLn('internal bug: filepos=$',
                          Int2Hex(programm_kopf[sektionen_uebersicht[sekt_z].index_].p_offset,8),
                          ' alignement mismatch to v_addr=$',
                          Int2Hex(programm_kopf[sektionen_uebersicht[sekt_z].index_].p_vaddr ,8));
                  RunError(1);
                end;

          BlockWrite_d2(d2_pos,speicher_tabelle[sektionen_uebersicht[sekt_z].index_]^,laenge);

        end;

    BlockWrite_d2(d2_pos,pk2,elf_kopf.e_phentsize*elf_kopf.e_phnum);

    d2^.Done
    // Writeln('Success.');
    // memleaks...
  end;

//

function RPos(C: Char; const S: string): LongInt;
begin
  Result := Length(S);
  while (Result > 0) and (S[Result] <> C) do Dec(Result);
end;

procedure show_usage;
  var
    z:word;
  begin
    WriteLn;
    WriteLn('Usage: PE2ELF [options ..] <infile>[.exe] [<outfile>]');
    WriteLn;
    WriteLn(' options:');
    WriteLn('  -v .. verbose');
    WriteLn('  -m1 smaller executable');
    WriteLn('  -m2 align to 4096 Byte');
    WriteLn('  -m3 put all uninitialized variables into executable');
//  WriteLn('  -h write section headers');
    WriteLn('  -i<program loader>');
    WriteLn('  -t:<OS/ABI>  valid values are:');
    for z:=Low(OSNames) to High(OSNames) do
      begin
        Write('     -t:',z);
        if z<10 then write(' ');
        Write(' or -t:',OSNames[z],Copy('                ',Length(OSNames[z]),25));
        if Odd(z) then WriteLn;
      end;
    if not Odd(High(OSNames)) then WriteLn;
    WriteLn('  -c:<target codepage> like IBM437, ISO-8859-1; default ISO-8859-1');
    Halt(1);
  end;

var
  param,z       :word;
  para          :string;
  ParaC         :char;
  kontrolle     :integer;

begin
  WriteLn('PE2ELF * Veit Kannegieser * 1999.07.28..',datum);

  // die Doku sagt '/usr/lib/libc.so.1'#0
  StrCopy(programmlader,'/lib/ld-linux.so.2');

  param:=1;
  repeat
    para:=ParamStr(param);
    if Pos('-',ParamStr(param))=0 then Break;

    Delete(para,1,1);
    if para='' then
      ParaC:=' '
    else
      begin
        ParaC:=UpCase(para[1]);
        Delete(para,1,1);
      end;

    if Pos(':',para)=1 then Delete(para,1,Length(':'));

    case ParaC of
      'V':
        verbose:=(para<>'-');

      'M':
        begin
          Val(para,memorymode,kontrolle);
          if (memorymode<0) or (memorymode>3) or (kontrolle<>0) then
            show_usage;
        end;

      'I':
        StrPCopy(programmlader,para);

      'H':
        include_section_header:=(para<>'-');

      'T':
        begin
          if para='' then show_usage;

          (* -t:9 ? *)
          Val(para[2],elf_kopf.osabi,kontrolle);
          if kontrolle=0 then para:='';

          (* -t:FreeBSD ? *)
          z:=Low(OSNames);
          while (z<=High(OSNames)) and (para<>'') do
            if Pos(para,OSNames[z])>0 then
              begin
                elf_kopf.osabi:=z;
                para:='';
              end
            else
              Inc(z);

          if para<>'' then show_usage;
        end;

      'C':resource_codepage:=para;

    else
      show_usage;
    end;

    Inc(param);
  until false;

  InFile := ParamStr(param);
  if InFile = '' then
    show_usage;

  Inc(param);

  if RPos('.', InFile) = 0 then InFile := InFile + '.exe';
  OutFile := ParamStr(param);
  if OutFile = '' then OutFile := InFile;
  Delete(Outfile, RPos('.', OutFile), 255);

  WriteLn('Converting PE file ''', InFile, ''' to ELF...');

  lade_pe;
  rechne;
  schreibe_elf;
end.

//elf doku seite 42: note
// nach vadr sortieren

