// used by PE2LE and PE2ELF
(*$Define PE2ELF_USE*)
(* $UnDef PE2ELF_USE*)
(*$Alignrec-*)
{&Use32+}
unit reskonv; (* Convert Resource(tree)s *)

(*$Define Use_Unicode*) (* include full codepage converter support *)

interface

uses
  Objects;

procedure konvertiere_resourcen(
  const quelle                          :pointer;
  const quell_laenge,quell_rva          :longint;
  const idata                           :pointer;
  const idata_laenge                    :longint;
  var ziel                              :pByteArray;
  var ziel_laenge                       :longint;
  const target_codepage                 :string);

// only used by PE2ELF
function search_resource_anchor(
  const daten                   :pByteArray;
  const daten_laenge            :longint):boolean;

// only used by PE2ELF
procedure patch_resource_anchor(
  var   daten                   :pByteArray;
  const resource_ofs            :longint;
  const resource_len            :longint);

// only used by PE2ELF
var
  resource_anchor_ofs           :longint=0;


implementation


uses
(*$IfDef Use_Unicode*)
  ntab,
  ztab,
  zst,
  UniDef,
(*$EndIf Use_Unicode*)
(*$IfNDef PE2ELF_USE*)
  pe2lespr,
  filter,
(*$EndIf PE2ELF_USE*)
  Dos;


const
  resourcetype_PE2XX_stringtable        =0;
  resourcetype_PE2XX_RCData             =1;
  resourcetype_PE2XX_NameString         =2; // Resource Name resource block, related resource will follow

  starting_name_id_for_stringnames      :Word=10001;

type
  (* pointer from code to resources (same in d32Res.pas/LnxRes.pas) *)
  dpmi32_resource_anchor_type   =
    packed record
      ptr_chain                 :Pointer;
      resource_size             :Longint;
      signature                 :array[0..3] of Char;
    end;

  (* resource block header *)
  dpmi32_resource_block_type    =
    packed record
      d32_resource_length       :Longint; (* size including this header *)
      d32_resource_type,
      d32_resource_id           :SmallWord;
    end;


  RESOURCE_DIR_ENTRIES_typ=
    record
      NAME_RVA_or_INTEGER_ID            :longint;
      DATA_ENTRY_RVA_or_SUBDIR_RVA      :longint; (* Bit 31=1 -> Subdir *)
    end;

  Resource_Directory_Table_typ=
    record
      RESOURCE_FLAGS                    :longint;
      TIME_DATE_STAMP                   :longint;
      MAJOR_VERSION                     :smallword;
      MINOR_VERSION                     :smallword;
      number_of_NAME_ENTRY              :smallword;
      number_of_ID_ENTRY                :smallword;
      RESOURCE_DIR_ENTRIES              :array[1..100000] of RESOURCE_DIR_ENTRIES_typ;
    end;

  Resource_Directory_Table_z_typ        =^Resource_Directory_Table_typ;

  Resource_Data_Entry=
    record
      DATA_RVA                          :longint;
      SIZE                              :longint;
      CODEPAGE                          :longint;
      RESERVED                          :longint;
    end;

  Resource_Data_Entry_z_typ             =^Resource_Data_Entry;


(*$IfDef Use_Unicode*)
(* Unicode converter tables *)
{&OrgName+}
procedure ntab_obj;external;
  {$L ntab32.obj}
procedure ztab_obj;external;
  {$L ztab32.obj}
{&OrgName-}

var
  unicode_to_cp_converter               :zeichensatzumformung_16_8;
(*$EndIf Use_Unicode*)

(*$IfDef Use_Unicode*)
procedure convert_16_8a(const u:pSmallWord;var s:Ansistring;lmax:Word);
  begin
    umformung_16l_8a(UniCharArray(u^),s,unicode_to_cp_converter,lmax);
  end;
(*$Else Use_Unicode*)
procedure convert_16_8a(const u:pSmallWord;var s:Ansistring;lmax:Word);
  var
    i           :word;
  begin
    (* this convert procedure is ok for 7 bit only.. *)
    s:='';
    for i:=1 to lmax do
      begin
        if u^=0 then Break;
        SetLength(s,i);
        s[i]:=Chr(u^);
        Inc(u);
      end;
  end;
(*$EndIf Use_Unicode*)



procedure konvertiere_resourcen(
  const quelle                          :pointer;
  const quell_laenge,quell_rva          :longint;
  const idata                           :pointer;
  const idata_laenge                    :longint;
  var ziel                              :pByteArray;
  var ziel_laenge                       :longint;
  const target_codepage                 :string);


  procedure schreibe(const a;laenge:longint);
    begin
      ReAllocMem(ziel,ziel_laenge+laenge);
      Move(a,ziel^[ziel_laenge],laenge);
      Inc(ziel_laenge,laenge);
    end;

  var
    resource_typid,resource_nameid      :smallword;
    resource_nameid_string              :ansistring;

    dpmi32_resource_block               :dpmi32_resource_block_type;

  procedure kopiere_resourcen(const r:Resource_Data_Entry_z_typ);
    var
      weiter                            :pointer;
      zaehler,
      string_laenge,
      lese_position,
      schreib_position                  :longint;
      temp                              :array[0..16*8000] of byte;
      resource_string                   :AnsiString;
    begin

      weiter:=Ptr(Ofs(quelle^)+r.DATA_RVA-quell_rva);
      dpmi32_resource_block.d32_resource_id:=resource_nameid;

      if resource_nameid_string<>'' then
        with dpmi32_resource_block do
          begin
            (* get an 'unique' number and place a number<->string association
               for the ID, and also for the following (RCData) block *)
            d32_resource_length:=SizeOf(dpmi32_resource_block)+2{SizeOf(string_laenge)}+Length(resource_nameid_string);
            d32_resource_type:=resourcetype_PE2XX_NameString;
            d32_resource_id:=starting_name_id_for_stringnames;
            Inc(starting_name_id_for_stringnames);
            schreibe(dpmi32_resource_block,SizeOf(dpmi32_resource_block));
            string_laenge:=Length(resource_nameid_string);
            schreibe(string_laenge,2{SizeOf(string_laenge)});
            schreibe(resource_nameid_string[1],Length(resource_nameid_string));
          end;

      if resource_typid=6 (* StringTable *) then
        begin
          dpmi32_resource_block.d32_resource_type:=resourcetype_PE2XX_stringtable;
          Dec(dpmi32_resource_block.d32_resource_id); (* $ff80 -> $0ff9 -> $0ff8 *)

          lese_position:=ofs(weiter^);
          schreib_position:=0;

          for zaehler:=0 to 15 do
            begin
              string_laenge:=memw[lese_position];
              inc(lese_position,2);
              convert_16_8a(pSmallWord(Ptr(lese_position)),
                            resource_string,
                            string_laenge);
              Move(string_laenge,temp[schreib_position],2{SizeOf(string_laenge)});
              Inc(schreib_position,2);
              Move(resource_string[1],temp[schreib_position],string_laenge);
              Inc(lese_position,2*string_laenge);
              Inc(schreib_position,string_laenge);

            end;
          dpmi32_resource_block.d32_resource_length:=SizeOf(dpmi32_resource_block)+schreib_position;
          schreibe(dpmi32_resource_block,SizeOf(dpmi32_resource_block));
          schreibe(temp,schreib_position);
        end
      else (* other binary/rcdata *)
        begin
          dpmi32_resource_block.d32_resource_type:=resourcetype_PE2XX_RCData;
          dpmi32_resource_block.d32_resource_length:=SizeOf(dpmi32_resource_block)+r.size;
          schreibe(dpmi32_resource_block,SizeOf(dpmi32_resource_block));
          schreibe(weiter^,r.size);
        end;
    end;

  procedure Language_baum(const l:Resource_Directory_Table_z_typ);
    var
      zaehler                           :longint;
      weiter                            :pointer;
    begin
      with l^ do
        for zaehler:=1 to number_of_NAME_ENTRY+number_of_ID_ENTRY do
          with RESOURCE_DIR_ENTRIES[zaehler] do
            begin
              if (zaehler<=number_of_NAME_ENTRY)
              or (NAME_RVA_or_INTEGER_ID<>0) then
                (*$IfDef PE2ELF_USE*)
                WriteLn('warning: no multilanguage support!');
                (*$Else PE2ELF_USE*)
                WriteLn(textz_warning_no_multilanguage_support^);
                (*$EndIf PE2ELF_USE*)
              weiter:=ptr(ofs(quelle^)+(DATA_ENTRY_RVA_or_SUBDIR_RVA and $7fffffff));
              if (DATA_ENTRY_RVA_or_SUBDIR_RVA and $80000000)=0 then
                kopiere_resourcen(Resource_Data_Entry_z_typ(weiter))
              else (* or flag them as error? *)
                Language_baum(Resource_Directory_Table_z_typ(weiter));
            end;
    end;

  procedure NameId_baum(const l:Resource_Directory_Table_z_typ);
    var
      zaehler                           :longint;
      weiter                            :pointer;
    begin
      with l^ do
        begin

          for zaehler:=1 to number_of_NAME_ENTRY+number_of_ID_ENTRY do
            with RESOURCE_DIR_ENTRIES[zaehler] do
              begin
                if zaehler<=number_of_NAME_ENTRY then
                  begin
                    resource_nameid:=0;
                    convert_16_8a(pSmallWord(Ptr(Ofs(quelle^)+2+(NAME_RVA_or_INTEGER_ID and $7fffffff))),
                                  resource_nameid_string,
                                  pSmallWord(Ptr(Ofs(quelle^)+0+(NAME_RVA_or_INTEGER_ID and $7fffffff)))^);
                  end
                else
                  begin
                    resource_nameid:=NAME_RVA_or_INTEGER_ID;
                    resource_nameid_string:='';
                  end;

                weiter:=Ptr(Ofs(quelle^)+(DATA_ENTRY_RVA_or_SUBDIR_RVA and $7fffffff));
                if (DATA_ENTRY_RVA_or_SUBDIR_RVA and $80000000)=0 then
                  kopiere_resourcen(Resource_Data_Entry_z_typ(weiter))
                else
                  Language_baum(Resource_Directory_Table_z_typ(weiter));
              end;
        end;
    end;

  procedure TypeId_baum(const l:Resource_Directory_Table_z_typ);
    var
      zaehler                           :longint;
      weiter                            :pointer;
    begin
      with l^ do
        for zaehler:=1 to number_of_NAME_ENTRY+number_of_ID_ENTRY do
          with RESOURCE_DIR_ENTRIES[zaehler] do
            begin

              if zaehler<=number_of_NAME_ENTRY then
                begin
                  (*$IfDef PE2ELF_USE*)
                  WriteLn('warning: no string type-id support!');
                  (*$Else PE2ELF_USE*)
                  writeln(textz_warning_no_string_typeid_support^);
                  (*$EndIf PE2ELF_USE*)
                  Continue;
                end;

              resource_typid:=NAME_RVA_or_INTEGER_ID;
              resource_nameid:=0;
              resource_nameid_string:='';
              weiter:=ptr(ofs(quelle^)+(DATA_ENTRY_RVA_or_SUBDIR_RVA and $7fffffff));
              if (DATA_ENTRY_RVA_or_SUBDIR_RVA and $80000000)=0 then
                kopiere_resourcen(Resource_Data_Entry_z_typ(weiter))
              else
                NameId_baum(Resource_Directory_Table_z_typ(weiter));
            end;
    end;


  var
    cp                  :word;
  begin
    if Assigned(quelle) then
      begin

        (*$IfDef Use_Unicode*)
        lade_speicher__cp_ntab(Addr(ntab_obj));
        lade_pakete_aus_dem_speicher(Addr(ztab_obj));
        berechne_umrechnungstabelle_16_8(umwandlung(target_codepage),unicode_to_cp_converter);
        (*$EndIf Use_Unicode*)

        TypeId_baum(Resource_Directory_Table_z_typ(quelle));

        (*$IfDef Use_Unicode*)
        freigeben_umrechnungstabelle_16_8(unicode_to_cp_converter);
        (*$EndIf Use_Unicode*)

      end;

    (*$IfNDef Use_PE2Elf*)
    if idata_laenge<>0 then
      begin
        dpmi32_resource_block.d32_resource_type:=2; (* import *)
        dpmi32_resource_block.d32_resource_id:=$ffff;
        dpmi32_resource_block.d32_resource_length:=SizeOf(dpmi32_resource_block)+idata_laenge;
        schreibe(dpmi32_resource_block,SizeOf(dpmi32_resource_block));
        schreibe(idata^,idata_laenge);
      end;
    (*$EndIf -Use_PE2Elf*)
  end;


function search_resource_anchor;
  var
    z1:longint;
  begin
    Result:=false;
    resource_anchor_ofs:=0;

    for z1:=0 to daten_laenge-Sizeof(dpmi32_resource_anchor_type) do
      // vielleicht zu langsam ..
      with dpmi32_resource_anchor_type(daten^[z1]) do
        if  (ptr_chain    =Ptr($b582cc18))
        and (resource_size=    $467a1d0e )
        and (signature    =    'RES?'    ) then
          begin
            resource_anchor_ofs:=z1;
            Result:=true;
            Exit;
          end;
  end;

procedure patch_resource_anchor;
  begin
    with dpmi32_resource_anchor_type(daten^[resource_anchor_ofs]) do
      begin
        ptr_chain    :=pointer(resource_ofs);
        resource_size:=resource_len;
        signature    :='RESO';
      end;
  end;

end.

