//
//                                                      
//     Virtual Pascal Runtime Library.  Version 2.1     
//     Win32 Tool Help Functions, types and definitions 
//     
//     Copyright (C) 2003 vpascal.com                   
//     VPC version by DataCompBoy [datacompboy@mail.ru] 
//     Copyright (C) 1994-1995 Microsoft Corporation    
//                                                      
//

unit tlhelp32;

interface

{&AlignRec+,StdCall+,SmartLink+,OrgName+}

uses
  Windows;

const
  MAX_MODULE_NAME32 = 255;

//****** Shapshot function **********************************************/

function CreateToolhelp32Snapshot(dwFlags: DWord; th32ProcessID: DWord): HANDLE;
//
// The th32ProcessID argument is only used if TH32CS_SNAPHEAPLIST or
// TH32CS_SNAPMODULE is specified. th32ProcessID == 0 means the current
// process.
//
// NOTE that all of the snapshots are global except for the heap and module
//  lists which are process specific. To enumerate the heap or module
//  state for all WIN32 processes call with TH32CS_SNAPALL and the
//  current process. Then for each process in the TH32CS_SNAPPROCESS
//  list that isn't the current process, do a call with just
//  TH32CS_SNAPHEAPLIST and/or TH32CS_SNAPMODULE.
//
// dwFlags
//
const   TH32CS_SNAPHEAPLIST = $00000001;
const   TH32CS_SNAPPROCESS  = $00000002;
const   TH32CS_SNAPTHREAD   = $00000004;
const   TH32CS_SNAPMODULE   = $00000008;
const   TH32CS_SNAPALL      = (TH32CS_SNAPHEAPLIST or TH32CS_SNAPPROCESS or TH32CS_SNAPTHREAD or TH32CS_SNAPMODULE);
const   TH32CS_INHERIT      = $80000000;
//
// Use CloseHandle to destroy the snapshot
//

//****** heap walking ***************************************************/

type LPHEAPLIST32 = ^THEAPLIST32;
     PHEAPLIST32 = ^THEAPLIST32;
     THEAPLIST32 = record
       dwSize        : DWord;
       th32ProcessID : DWord; // owning process
       th32HeapID    : DWord; // heap (in owning process's context!)
       dwFlags       : DWord;
     end;

//
// dwFlags
//
const HF32_DEFAULT = 1;  // process's default heap
const HF32_SHARED  = 2;  // is shared heap

function Heap32ListFirst(hSnapshot: HANDLE; lphl: LPHEAPLIST32):BOOL;
function Heap32ListNext (hSnapshot: HANDLE; lphl: LPHEAPLIST32):BOOL;

type LPHEAPENTRY32 = ^THEAPENTRY32;
     PHEAPENTRY32 = ^THEAPENTRY32;
     THEAPENTRY32 = record
       dwSize       : DWord ;
       hHandle      : HANDLE; // Handle of this heap block
       dwAddress    : DWord ;   // Linear address of start of block
       dwBlockSize  : DWord ; // Size of block in bytes
       dwFlags      : DWord ;
       dwLockCount  : DWord ;
       dwResvd      : DWord ;
       th32ProcessID: DWord ;   // owning process
       th32HeapID   : DWord ;      // heap block is in
     end;

//
// dwFlags
//
const LF32_FIXED    = $00000001;
const LF32_FREE     = $00000002;
const LF32_MOVEABLE = $00000004;

function Heap32First(lphe: LPHEAPENTRY32; th32ProcessID: DWord;
                     th32HeapID: DWord): BOOL;
function Heap32Next(lphe: LPHEAPENTRY32): BOOL;
function Toolhelp32ReadProcessMemory(th32ProcessID: DWord;
                    lpBaseAddress       : LPVOID;
                    lpBuffer            : LPVOID;
                    cbRead              : DWord;
                    lpNumberOfBytesRead : LPDWord): BOOL;

//***** Process walking *************************************************/

type LPPROCESSENTRY32 = ^TPROCESSENTRY32;
     PPROCESSENTRY32 = ^TPROCESSENTRY32;
     TPROCESSENTRY32 = record
       dwSize              : DWord;
       cntUsage            : DWord;
       th32ProcessID       : DWord; // this process
       th32DefaultHeapID   : DWord;
       th32ModuleID        : DWord; // associated exe
       cntThreads          : DWord;
       th32ParentProcessID : DWord; // this process's parent process
       pcPriClassBase      : LONG ; // Base priority of process's threads
       dwFlags             : DWord;
       szExeFile : array [0..Max_Path-1] of char;  // Path
     end;

function Process32First(hSnapshot: HANDLE; lppe: LPPROCESSENTRY32): BOOL;
function Process32Next (hSnapshot: HANDLE; lppe: LPPROCESSENTRY32): BOOL;

//***** Thread walking **************************************************/

type LPTHREADENTRY32 = ^TTHREADENTRY32;
     PTHREADENTRY32 = ^TTHREADENTRY32;
     TTHREADENTRY32 = record
       dwSize             : DWord;
       cntUsage           : DWord;
       th32ThreadID       : DWord; // this thread
       th32OwnerProcessID : DWord; // Process this thread is associated with
       tpBasePri          : LONG ;
       tpDeltaPri         : LONG ;
       dwFlags            : DWord;
     end;

function Thread32First(hSnapshot: HANDLE; lpte: LPTHREADENTRY32): BOOL;
function Thread32Next (hSnapshot: HANDLE; lpte: LPTHREADENTRY32): BOOL;

//***** Module walking *************************************************/

type LPMODULEENTRY32 = ^TMODULEENTRY32;
     PMODULEENTRY32 = ^TMODULEENTRY32;
     TMODULEENTRY32 = record
       dwSize        : DWord  ;
       th32ModuleID  : DWord  ; // This module
       th32ProcessID : DWord  ; // owning process
       GlblcntUsage  : DWord  ; // Global usage count on the module
       ProccntUsage  : DWord  ; // Module usage count in th32ProcessID's context
       modBaseAddr   : PByte  ; // Base address of module in th32ProcessID's context
       modBaseSize   : DWord  ; // Size in bytes of module starting at modBaseAddr
       hModule       : HMODULE; // The hModule of this module in th32ProcessID's context
       szModule      : array [0..MAX_MODULE_NAME32] of char;
       szExePath     : array [0..Max_Path-1] of char;
     end;

//
// NOTE CAREFULLY that the modBaseAddr and hModule fields are valid ONLY
// in th32ProcessID's process context.
//

function Module32First(hSnapshot: HANDLE; lpme: LPMODULEENTRY32): BOOL;
function Module32Next (hSnapshot: HANDLE; lpme: LPMODULEENTRY32): BOOL;

implementation

function CreateToolhelp32Snapshot   ; external;
function Heap32ListFirst            ; external; 
function Heap32ListNext             ; external; 
function Heap32First                ; external; 
function Heap32Next                 ; external; 
function Toolhelp32ReadProcessMemory; external; 
function Process32First             ; external; 
function Process32Next              ; external; 
function Thread32First              ; external; 
function Thread32Next               ; external; 
function Module32First              ; external; 
function Module32Next               ; external; 

end.
