//
//                                                      
//     Virtual Pascal Runtime Library.  Version 2.1     
//     Windows Cryptographic API Prototypes             
//     
//     Copyright (C) 2003 vpascal.com                   
//     VPC version by DataCompBoy [datacompboy@mail.ru] 
//     Copyright (C) 1992-1996 Microsoft Corporation    
//                                                      
//

unit WinCrypt;

interface
{&Stdcall+,Orgname+}

uses
  Windows;

//
// Algorithm IDs and Flags
//

// ALG_ID crackers
// GET_ALG_CLASS(x)                (x & (7 << 13))
function GET_ALG_CLASS(x: DWORD): DWORD; inline;
begin GET_ALG_CLASS:=x and (7 shl 13) end;

// GET_ALG_TYPE(x)                 (x & (15 << 9))
function GET_ALG_TYPE(x: DWORD): DWORD; inline;
begin GET_ALG_TYPE:=x and (15 shl 9) end;

// GET_ALG_SID(x)                  (x & (511))
function GET_ALG_SID(x: DWORD): DWORD; inline;
begin GET_ALG_SID:=x and (511) end;

// Algorithm classes
const
  ALG_CLASS_ANY          = (0);
  ALG_CLASS_SIGNATURE    = (1 shl 13);
  ALG_CLASS_MSG_ENCRYPT  = (2 shl 13);
  ALG_CLASS_DATA_ENCRYPT = (3 shl 13);
  ALG_CLASS_HASH         = (4 shl 13);
  ALG_CLASS_KEY_EXCHANGE = (5 shl 13);

// Algorithm types
const
  ALG_TYPE_ANY           = (0);
  ALG_TYPE_DSS           = (1 shl 9);
  ALG_TYPE_RSA           = (2 shl 9);
  ALG_TYPE_BLOCK         = (3 shl 9);
  ALG_TYPE_STREAM        = (4 shl 9);

// Generic sub-ids
const
  ALG_SID_ANY            = (0);

// Some RSA sub-ids
const
  ALG_SID_RSA_ANY        = 0;
  ALG_SID_RSA_PKCS       = 1;
  ALG_SID_RSA_MSATWORK   = 2;
  ALG_SID_RSA_ENTRUST    = 3;
  ALG_SID_RSA_PGP        = 4;

// Some DSS sub-ids
const
  ALG_SID_DSS_ANY        = 0;
  ALG_SID_DSS_PKCS       = 1;
  ALG_SID_DSS_DMS        = 2;

// Block cipher sub ids
// DES sub_ids
const
  ALG_SID_DES            = 1;
  ALG_SID_3DES           = 3;
  ALG_SID_DESX           = 4;
  ALG_SID_IDEA           = 5;
  ALG_SID_CAST           = 6;
  ALG_SID_SAFERSK64      = 7;
  ALD_SID_SAFERSK128     = 8;

// KP_MODE
const
  CRYPT_MODE_CBCI        = 6;     // ANSI CBC Interleaved
  CRYPT_MODE_CFBP        = 7;     // ANSI CFB Pipelined
  CRYPT_MODE_OFBP        = 8;     // ANSI OFB Pipelined
  CRYPT_MODE_CBCOFM      = 9;     // ANSI CBC + OF Masking
  CRYPT_MODE_CBCOFMI     = 10;    // ANSI CBC + OFM Interleaved

// RC2 sub-ids
const
  ALG_SID_RC2            = 2;

// Stream cipher sub-ids
const
  ALG_SID_RC4            = 1;
  ALG_SID_SEAL           = 2;

// Hash sub ids
const
  ALG_SID_MD2            = 1;
  ALG_SID_MD4            = 2;
  ALG_SID_MD5            = 3;
  ALG_SID_SHA            = 4;
  ALG_SID_MAC            = 5;
  ALG_SID_RIPEMD         = 6;
  ALG_SID_RIPEMD160      = 7;
  ALG_SID_SSL3SHAMD5     = 8;


// Our silly example sub-id
const
  ALG_SID_EXAMPLE        = 80;

type ALG_ID = DWORD;

// algorithm identifier definitions
const
  CALG_MD2      = (ALG_CLASS_HASH or ALG_TYPE_ANY or ALG_SID_MD2);
  CALG_MD4      = (ALG_CLASS_HASH or ALG_TYPE_ANY or ALG_SID_MD4);
  CALG_MD5      = (ALG_CLASS_HASH or ALG_TYPE_ANY or ALG_SID_MD5);
  CALG_SHA      = (ALG_CLASS_HASH or ALG_TYPE_ANY or ALG_SID_SHA);
  CALG_MAC      = (ALG_CLASS_HASH or ALG_TYPE_ANY or ALG_SID_MAC);
  CALG_RSA_SIGN = (ALG_CLASS_SIGNATURE or ALG_TYPE_RSA or ALG_SID_RSA_ANY);
  CALG_DSS_SIGN = (ALG_CLASS_SIGNATURE or ALG_TYPE_DSS or ALG_SID_DSS_ANY);
  CALG_RSA_KEYX = (ALG_CLASS_KEY_EXCHANGE or ALG_TYPE_RSA or ALG_SID_RSA_ANY);
  CALG_DES      = (ALG_CLASS_DATA_ENCRYPT or ALG_TYPE_BLOCK or ALG_SID_DES);
  CALG_RC2      = (ALG_CLASS_DATA_ENCRYPT or ALG_TYPE_BLOCK or ALG_SID_RC2);
  CALG_RC4      = (ALG_CLASS_DATA_ENCRYPT or ALG_TYPE_STREAM or ALG_SID_RC4);
  CALG_SEAL     = (ALG_CLASS_DATA_ENCRYPT or ALG_TYPE_STREAM or ALG_SID_SEAL);

type
  TVTableProvStruc = record
    Version: DWORD;
    FuncVerifyImage: FARPROC;
    FuncReturnhWnd: FARPROC;
  end;
  PVTableProvStruc = ^TVTableProvStruc;

type HCRYPTPROV= DWORD;
type HCRYPTKEY = DWORD;
type HCRYPTHASH= DWORD;

// dwFlags definitions for CryptAquireContext
const
  CRYPT_VERIFYCONTEXT   =  $F0000000;
  CRYPT_NEWKEYSET       =  $8;
  CRYPT_DELETEKEYSET    =  $10;

  // dwFlag definitions for CryptGenKey
  CRYPT_EXPORTABLE      =  $00000001;
  CRYPT_USER_PROTECTED  =  $00000002;
  CRYPT_CREATE_SALT     =  $00000004;
  CRYPT_UPDATE_KEY      =  $00000008;

  // exported key blob definitions
  SIMPLEBLOB            =  $1;
  PUBLICKEYBLOB         =  $6;
  PRIVATEKEYBLOB        =  $7;

  AT_KEYEXCHANGE        =  1;
  AT_SIGNATURE          =  2;

  CRYPT_USERDATA        =  1;

// dwParam
const
  KP_IV                 =  1;       // Initialization vector
  KP_SALT               =  2;       // Salt value
  KP_PADDING            =  3;       // Padding values
  KP_MODE               =  4;       // Mode of the cipher
  KP_MODE_BITS          =  5;       // Number of bits to feedback
  KP_PERMISSIONS        =  6;       // Key permissions DWORD
  KP_ALGID              =  7;       // Key algorithm
  KP_BLOCKLEN           =  8;       // Block size of the cipher

// KP_PADDING
const
  PKCS5_PADDING         =  1;       // PKCS 5 (sec 6.2) padding method

// KP_MODE
const
  CRYPT_MODE_CBC        =  1;       // Cipher block chaining
  CRYPT_MODE_ECB        =  2;       // Electronic code book
  CRYPT_MODE_OFB        =  3;       // Output feedback mode
  CRYPT_MODE_CFB        =  4;       // Cipher feedback mode
  CRYPT_MODE_CTS        =  5;       // Ciphertext stealing mode

// KP_PERMISSIONS
const
  CRYPT_ENCRYPT         =  $0001;  // Allow encryption
  CRYPT_DECRYPT         =  $0002;  // Allow decryption
  CRYPT_EXPORT          =  $0004;  // Allow key to be exported
  CRYPT_READ            =  $0008;  // Allow parameters to be read
  CRYPT_WRITE           =  $0010;  // Allow parameters to be set
  CRYPT_MAC             =  $0020;  // Allow MACs to be used with key

  HP_ALGID              =  $0001;  // Hash algorithm
  HP_HASHVAL            =  $0002;  // Hash value
  HP_HASHSIZE           =  $0004;  // Hash value size


  CRYPT_FAILED          =  FALSE;
  CRYPT_SUCCEED         =  TRUE;

// RCRYPT_SUCCEEDED(rt)     ((rt) == CRYPT_SUCCEED)
function RCRYPT_SUCCEEDED(rt: DWORD): BOOL; inline;
begin RCRYPT_SUCCEEDED:=BOOL(rt)=CRYPT_SUCCEED end;

// RCRYPT_FAILED(rt)        ((rt) == CRYPT_FAILED)
function RCRYPT_FAILED(rt: DWORD): BOOL; inline;
begin RCRYPT_FAILED:=BOOL(rt)=CRYPT_FAILED end;

//
// CryptGetProvParam
//
const
  PP_ENUMALGS           =  1;
  PP_ENUMCONTAINERS     =  2;
  PP_IMPTYPE            =  3;
  PP_NAME               =  4;
  PP_VERSION            =  5;
  PP_CONTAINER          =  6;

  CRYPT_FIRST           =  1;
  CRYPT_NEXT            =  2;

  CRYPT_IMPL_HARDWARE   =  1;
  CRYPT_IMPL_SOFTWARE   =  2;
  CRYPT_IMPL_MIXED      =  3;
  CRYPT_IMPL_UNKNOWN    =  4;

//
// CryptSetProvParam
//
const
  PP_CLIENT_HWND        =  1;

  PROV_RSA_FULL         =  1;
  PROV_RSA_SIG          =  2;
  PROV_DSS              =  3;
  PROV_FORTEZZA         =  4;
  PROV_MS_EXCHANGE      =  5;
  PROV_SSL              =  6;

//
//STT defined Providers
//
const
  PROV_STT_MER                  =  7;
  PROV_STT_ACQ                  =  8;
  PROV_STT_BRND                 =  9;
  PROV_STT_ROOT                 =  10;
  PROV_STT_ISS                  =  11;

const
  MS_DEF_PROV_A      = 'Microsoft Base Cryptographic Provider v1.0';
  MS_DEF_PROV_W      = #0'M'#0'i'#0'c'#0'r'#0'o'#0's'#0'o'#0'f'#0't'#0' '#0'B'#0'a'#0's'#0'e'#0' '#0'C'#0'r'#0'y'#0'p'#0't'#0'o'#0'g'#0'r'#0'a'#0'p'#0'h'#0'i'#0'c'#0' '#0'P'#0'r'#0'o'#0'v'#0'i'#0'd'#0'e'#0'r'#0' '#0'v'#0'1'#0'.'#0'0';
  MS_DEF_PROV        = MS_DEF_PROV_A;

  MAXUIDLEN          = 64;

  CUR_BLOB_VERSION   =  2;

type
  TPROV_ENUMALGS= record
    aiAlgid: ALG_ID;
    dwBitLen: DWORD;
    dwNameLen:DWORD;
    szName: array[0..19] of char;
  end;

type
  PUBLICKEYSTRUC=record
    bType: byte;
    bVersion: byte;
    reserved: smallword;
    aiKeyAlg: ALG_ID;
  end;

type
  RSAPUBKEY = record
    magic:DWORD;                  // Has to be RSA1
    bitlen:DWORD;                 // # of bits in modulus
    pubexp:DWORD;                 // public exponent
                                  // Modulus data follows
  end;

 /////////////////
 //  Error codes
 /////////////////
const
  //
  // MessageId: NTE_BAD_UID
  //
  // MessageText:
  //
  //  Bad UID.
  //
  NTE_BAD_UID                     = ($80090001);

  //
  // MessageId: NTE_BAD_HASH
  //
  // MessageText:
  //
  //  Bad Hash.
  //
  NTE_BAD_HASH                    = ($80090002);

  //
  // MessageId: NTE_BAD_KEY
  //
  // MessageText:
  //
  //  Bad Key.
  //
  NTE_BAD_KEY                     = ($80090003);

  //
  // MessageId: NTE_BAD_LEN
  //
  // MessageText:
  //
  //  Bad Length.
  //
  NTE_BAD_LEN                     = ($80090004);

  //
  // MessageId: NTE_BAD_DATA
  //
  // MessageText:
  //
  //  Bad Data.
  //
  NTE_BAD_DATA                    = ($80090005);

  //
  // MessageId: NTE_BAD_SIGNATURE
  //
  // MessageText:
  //
  //  Invalid Signature.
  //
  NTE_BAD_SIGNATURE                = ($80090006);

  //
  // MessageId: NTE_BAD_VER
  //
  // MessageText:
  //
  //  Bad Version of provider.
  //
  NTE_BAD_VER                      = ($80090007);

  //
  // MessageId: NTE_BAD_ALGID
  //
  // MessageText:
  //
  //  Invalid algorithm specified.
  //
  NTE_BAD_ALGID                    = ($80090008);

  //
  // MessageId: NTE_BAD_FLAGS
  //
  // MessageText:
  //
  //  Invalid flags specified.
  //
  NTE_BAD_FLAGS                    = ($80090009);

  //
  // MessageId: NTE_BAD_TYPE
  //
  // MessageText:
  //
  //  Invalid type specified.
  //
  NTE_BAD_TYPE                     = ($8009000A);

  //
  // MessageId: NTE_BAD_KEY_STATE
  //
  // MessageText:
  //
  //  Key not valid for use in specified state.
  //
  NTE_BAD_KEY_STATE                = ($8009000B);

  //
  // MessageId: NTE_BAD_HASH_STATE
  //
  // MessageText:
  //
  //  Hash not valid for use in specified state.
  //
  NTE_BAD_HASH_STATE               = ($8009000C);

  //
  // MessageId: NTE_NO_KEY
  //
  // MessageText:
  //
  //  Key does not exist.
  //
  NTE_NO_KEY                       = ($8009000D);

  //
  // MessageId: NTE_NO_MEMORY
  //
  // MessageText:
  //
  //  Insufficient memory available for the operation.
  //
  NTE_NO_MEMORY                    = ($8009000E);

  //
  // MessageId: NTE_EXISTS
  //
  // MessageText:
  //
  //  Object already exists.
  //
  NTE_EXISTS                       = ($8009000F);

  //
  // MessageId: NTE_PERM
  //
  // MessageText:
  //
  //  Access denied.
  //
  NTE_PERM                         = ($80090010);

  //
  // MessageId: NTE_NOT_FOUND
  //
  // MessageText:
  //
  //  Object was not found.
  //
  NTE_NOT_FOUND                    = ($80090011);

  //
  // MessageId: NTE_DOUBLE_ENCRYPT
  //
  // MessageText:
  //
  //  Data already encrypted.
  //
  NTE_DOUBLE_ENCRYPT               = ($80090012);

  //
  // MessageId: NTE_BAD_PROVIDER
  //
  // MessageText:
  //
  //  Invalid provider specified.
  //
  NTE_BAD_PROVIDER                 = ($80090013);

  //
  // MessageId: NTE_BAD_PROV_TYPE
  //
  // MessageText:
  //
  //  Invalid provider type specified.
  //
  NTE_BAD_PROV_TYPE                = ($80090014);

  //
  // MessageId: NTE_BAD_PUBLIC_KEY
  //
  // MessageText:
  //
  //  Provider's public key is invalid.
  //
  NTE_BAD_PUBLIC_KEY               = ($80090015);

  //
  // MessageId: NTE_BAD_KEYSET
  //
  // MessageText:
  //
  //  Keyset does not exist
  //
  NTE_BAD_KEYSET                   = ($80090016);

  //
  // MessageId: NTE_PROV_TYPE_NOT_DEF
  //
  // MessageText:
  //
  //  Provider type not defined.
  //
  NTE_PROV_TYPE_NOT_DEF            = ($80090017);

  //
  // MessageId: NTE_PROV_TYPE_ENTRY_BAD
  //
  // MessageText:
  //
  //  Provider type as registered is invalid.
  //
  NTE_PROV_TYPE_ENTRY_BAD          = ($80090018);

  //
  // MessageId: NTE_KEYSET_NOT_DEF
  //
  // MessageText:
  //
  //  The keyset is not defined.
  //
  NTE_KEYSET_NOT_DEF               = ($80090019);

  //
  // MessageId: NTE_KEYSET_ENTRY_BAD
  //
  // MessageText:
  //
  //  Keyset as registered is invalid.
  //
  NTE_KEYSET_ENTRY_BAD             = ($8009001A);

  //
  // MessageId: NTE_PROV_TYPE_NO_MATCH
  //
  // MessageText:
  //
  //  Provider type does not match registered value.
  //
  NTE_PROV_TYPE_NO_MATCH           = ($8009001B);

  //
  // MessageId: NTE_SIGNATURE_FILE_BAD
  //
  // MessageText:
  //
  //  The digital signature file is corrupt.
  //
  NTE_SIGNATURE_FILE_BAD           = ($8009001C);

  //
  // MessageId: NTE_PROVIDER_DLL_FAIL
  //
  // MessageText:
  //
  //  Provider DLL failed to initialize correctly.
  //
  NTE_PROVIDER_DLL_FAIL            = ($8009001D);

  //
  // MessageId: NTE_PROV_DLL_NOT_FOUND
  //
  // MessageText:
  //
  //  Provider DLL could not be found.
  //
  NTE_PROV_DLL_NOT_FOUND           = ($8009001E);

  //
  // MessageId: NTE_BAD_KEYSET_PARAM
  //
  // MessageText:
  //
  //  The Keyset parameter is invalid.
  //
  NTE_BAD_KEYSET_PARAM             = ($8009001F);

  //
  // MessageId: NTE_FAIL
  //
  // MessageText:
  //
  //  An internal error occurred.
  //
  NTE_FAIL                         = ($80090020);

  //
  // MessageId: NTE_SYS_ERR
  //
  // MessageText:
  //
  //  A base error occurred.
  //
  NTE_SYS_ERR                      = ($80090021);

  NTE_OP_OK = 0;



function CryptAcquireContext( //U
    var phProv: HCRYPTPROV;
    pszContainer: LPCSTR;
    pszProvider: LPCSTR;
    dwProvType: DWORD;
    dwFlags: DWORD): BOOL; stdcall;

function CryptAcquireContextA(
    var phProv: HCRYPTPROV;
    pszContainer: LPCSTR;
    pszProvider: LPCSTR;
    dwProvType: DWORD;
    dwFlags: DWORD): BOOL; stdcall;

function CryptAcquireContextW(
    var phProv: HCRYPTPROV;
    pszContainer: LPCSTR;
    pszProvider: LPCSTR;
    dwProvType: DWORD;
    dwFlags: DWORD): BOOL; stdcall;

function CryptReleaseContext(
    hProv: HCRYPTPROV;
    dwFlags: DWORD): BOOL; stdcall;

function CryptGenKey(
    hProv: HCRYPTPROV;
    Algid: ALG_ID;
    dwFlags: DWORD;
    var phKey: HCRYPTKEY): BOOL; stdcall;

function CryptDeriveKey(
    hProv: HCRYPTPROV;
    Algid: ALG_ID;
    hBaseData: HCRYPTHASH;
    dwFlags: DWORD;
    var phKey: HCRYPTKEY): BOOL; stdcall;


function CryptDestroyKey(
    hKey: HCRYPTKEY): BOOL; stdcall;

function CryptSetKeyParam(
    hKey: HCRYPTKEY;
    dwParam: DWORD;
    var pbData;
    dwFlags: DWORD): BOOL; stdcall;

function CryptGetKeyParam(
    hKey: HCRYPTKEY;
    dwParam: DWORD;
    var pbData;
    var pdwDataLen: DWORD;
    dwFlags: DWORD): BOOL; stdcall;

function  CryptSetHashParam(
    hHash: HCRYPTHASH;
    dwParam: DWORD;
    var pbData;
    dwFlags: DWORD): BOOL; stdcall;

function CryptGetHashParam(
    hHash: HCRYPTHASH;
    dwParam: DWORD;
    var pbData;
    var pdwDataLen: DWORD;
    dwFlags: DWORD): BOOL; stdcall;

function CryptSetProvParam(
    hProv: HCRYPTPROV;
    dwParam: DWORD;
    var pbData;
    dwFlags: DWORD): BOOL; stdcall;

function CryptGetProvParam(
    hProv: HCRYPTPROV;
    dwParam: DWORD;
    var pbData;
    var pdwDataLen: DWORD;
    dwFlags: DWORD): BOOL; stdcall;

function CryptGenRandom(
    hProv: HCRYPTPROV;
    dwLen: DWORD;
    var pbBuffer): BOOL; stdcall;

function CryptGetUserKey(
    hProv: HCRYPTPROV;
    dwKeySpec: DWORD;
    var phUserKey: HCRYPTKEY): BOOL; stdcall;

function CryptExportKey(
    hKey: HCRYPTKEY;
    hExpKey: HCRYPTKEY;
    dwBlobType: DWORD;
    dwFlags: DWORD;
    var pbData;
    var pdwDataLen: DWORD): BOOL; stdcall;

function CryptImportKey(
    hProv: HCRYPTPROV;
    CONST pbData;
    dwDataLen: DWORD;
    hPubKey: HCRYPTKEY;
    dwFlags: DWORD;
    var phKey: HCRYPTKEY): BOOL; stdcall;

function CryptEncrypt(
    hKey: HCRYPTKEY;
    hHash: HCRYPTHASH;
    Final: longbool;
    dwFlags: DWORD;
    var pbData;
    var pdwDataLen: DWORD;
    dwBufLen: DWORD): BOOL; stdcall;

function CryptDecrypt(
    hKey: HCRYPTKEY;
    hHash: HCRYPTHASH;
    Final: longbool;
    dwFlags: DWORD;
    var pbData;
    var pdwDataLen: DWORD): BOOL; stdcall;

function CryptCreateHash(
    hProv: HCRYPTPROV;
    Algid: ALG_ID;
    hKey: HCRYPTKEY;
    dwFlags: DWORD;
    var phHash: HCRYPTHASH): BOOL; stdcall;

function CryptHashData(
    hHash: HCRYPTHASH;
    CONST pbData;
    dwDataLen: DWORD;
    dwFlags: DWORD): BOOL; stdcall;

function CryptHashSessionKey(
    hHash: HCRYPTHASH;
    hKey: HCRYPTKEY;
    dwFlags: DWORD): BOOL; stdcall;

function CryptDestroyHash(
    hHash: HCRYPTHASH): BOOL; stdcall;

function CryptSignHash( //U
    hHash: HCRYPTHASH;
    dwKeySpec: DWORD;
    sDescription: LPCSTR;
    dwFlags: DWORD;
    var pbSignature;
    var pdwSigLen: DWORD): BOOL; stdcall;

function CryptSignHashA( //U
    hHash: HCRYPTHASH;
    dwKeySpec: DWORD;
    sDescription: LPCSTR;
    dwFlags: DWORD;
    var pbSignature;
    var pdwSigLen: DWORD): BOOL; stdcall;

function CryptSignHashW( //U
    hHash: HCRYPTHASH;
    dwKeySpec: DWORD;
    sDescription: LPCSTR;
    dwFlags: DWORD;
    var pbSignature;
    var pdwSigLen: DWORD): BOOL; stdcall;

function CryptVerifySignature( //U
    hHash: HCRYPTHASH;
    CONST pbSignature;
    dwSigLen: DWORD;
    hPubKey: HCRYPTKEY;
    sDescription: LPCSTR;
    dwFlags: DWORD): BOOL; stdcall;

function CryptVerifySignatureA( //U
    hHash: HCRYPTHASH;
    CONST pbSignature;
    dwSigLen: DWORD;
    hPubKey: HCRYPTKEY;
    sDescription: LPCSTR;
    dwFlags: DWORD): BOOL; stdcall;

function CryptVerifySignatureW( //U
    hHash: HCRYPTHASH;
    CONST pbSignature;
    dwSigLen: DWORD;
    hPubKey: HCRYPTKEY;
    sDescription: LPCSTR;
    dwFlags: DWORD): BOOL; stdcall;

function CryptSetProvider( //U
    pszProvName: LPCSTR;
    dwProvType: DWORD): BOOL; stdcall;

function CryptSetProviderA( //U
    pszProvName: LPCSTR;
    dwProvType: DWORD): BOOL; stdcall;

function CryptSetProviderW( //U
    pszProvName: LPCSTR;
    dwProvType: DWORD): BOOL; stdcall;

implementation

function CryptAcquireContext  ; external;
function CryptAcquireContextA ; external;
function CryptAcquireContextW ; external;
function CryptReleaseContext  ; external;
function CryptGenKey          ; external;
function CryptDeriveKey       ; external;
function CryptDestroyKey      ; external;
function CryptSetKeyParam     ; external;
function CryptGetKeyParam     ; external;
function CryptSetHashParam    ; external;
function CryptGetHashParam    ; external;
function CryptSetProvParam    ; external;
function CryptGetProvParam    ; external;
function CryptGenRandom       ; external;
function CryptGetUserKey      ; external;
function CryptExportKey       ; external;
function CryptImportKey       ; external;
function CryptEncrypt         ; external;
function CryptDecrypt         ; external;
function CryptCreateHash      ; external;
function CryptHashData        ; external;
function CryptHashSessionKey  ; external;
function CryptDestroyHash     ; external;
function CryptSignHash        ; external;
function CryptSignHashA       ; external;
function CryptSignHashW       ; external;
function CryptVerifySignature ; external;
function CryptVerifySignatureA; external;
function CryptVerifySignatureW; external;
function CryptSetProvider     ; external;
function CryptSetProviderA    ; external;
function CryptSetProviderW    ; external;

end.
